#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    jinja-based :class:`~nunavut.generators.AbstractGenerator` implementation.
"""

import datetime
import io
import logging
import pathlib
import re
import shutil
import typing

import nunavut.generators
import nunavut.lang
import nunavut.postprocessors
import pydsdl
from nunavut._utilities import YesNoDefault
from yaml import Dumper as YamlDumper
from yaml import dump as yaml_dump

from .environment import CodeGenEnvironment
from .jinja2 import Template
from .loaders import DEFAULT_TEMPLATE_PATH, TEMPLATE_SUFFIX, DSDLTemplateLoader

logger = logging.getLogger(__name__)

# +---------------------------------------------------------------------------+
# | JINJA : CodeGenerator
# +---------------------------------------------------------------------------+


class CodeGenerator(nunavut.generators.AbstractGenerator):
    """
    Abstract base class for all Generators that build source code using Jinja templates.

    :param nunavut.Namespace namespace:    The top-level namespace to generates code
                                           at and from.
    :param YesNoDefault generate_namespace_types:  Set to YES to emit files for namespaces.
                                           NO will suppress namespace file generation and DEFAULT will
                                           use the language's preference.
    :param templates_dir:                  Directories containing jinja templates. These will be available along
                                           with any built-in templates provided by the target language. The templates
                                           at these paths will take precedence masking any built-in templates
                                           where the names are the same. See :class:`jinja2.ChoiceLoader` for rules
                                           on the lookup hierarchy.
    :type templates_dir: typing.Optional[typing.Union[pathlib.Path,typing.List[pathlib.Path]]]
    :param bool followlinks:               If True then symbolic links will be followed when
                                           searching for templates.
    :param bool trim_blocks:               If this is set to True the first newline after a
                                           block is removed (block, not variable tag!).
    :param bool lstrip_blocks:             If this is set to True leading spaces and tabs
                                           are stripped from the start of a line to a block.
                                           Defaults to False.
    :param typing.Dict[str, typing.Callable] additional_filters: typing.Optional jinja filters to add to the
                                           global environment using the key as the filter name
                                           and the callable as the filter.
    :param typing.Dict[str, typing.Callable] additional_tests: typing.Optional jinja tests to add to the
                                           global environment using the key as the test name
                                           and the callable as the test.
    :param typing.Dict[str, typing.Any] additional_globals: typing.Optional objects to add to the template
                                            environment globals collection.
    :param post_processors: A list of :class:`nunavut.postprocessors.PostProcessor`
    :type post_processors: typing.Optional[typing.List[nunavut.postprocessors.PostProcessor]]
    :param builtin_template_path: If provided overrides the folder name under which built-in templates are loaded from
                                            within a target language's package (i.e. ignored if no target language is
                                            specified). For example, if the target language is ``c`` and this parameter
                                            was set to ``foo`` then built-in templates would be loaded from
                                            ``nunavut.lang.c.foo``.
    :raises RuntimeError: If any additional filter or test attempts to replace a built-in
                          or otherwise already defined filter or test.
    """

    @staticmethod
    def __augment_post_processors_with_ln_limit_empty_lines(
        post_processors: typing.Optional[typing.List["nunavut.postprocessors.PostProcessor"]], limit_empty_lines: int
    ) -> typing.List["nunavut.postprocessors.PostProcessor"]:
        """
        Subroutine of _handle_post_processors method.
        """
        from nunavut.postprocessors import LimitEmptyLines

        if post_processors is None:
            post_processors = [LimitEmptyLines(limit_empty_lines)]
        else:
            found_pp = False
            for pp in post_processors:
                if isinstance(pp, LimitEmptyLines):
                    found_pp = True
                    break
            if not found_pp:
                post_processors.append(LimitEmptyLines(limit_empty_lines))
        return post_processors

    @staticmethod
    def __augment_post_processors_with_ln_trim_trailing_whitespace(
        post_processors: typing.Optional[typing.List["nunavut.postprocessors.PostProcessor"]],
    ) -> typing.List["nunavut.postprocessors.PostProcessor"]:
        """
        Subroutine of _handle_post_processors method.
        """
        from nunavut.postprocessors import TrimTrailingWhitespace

        if post_processors is None:
            post_processors = [TrimTrailingWhitespace()]
        else:
            found_pp = False
            for pp in post_processors:
                if isinstance(pp, TrimTrailingWhitespace):
                    found_pp = True
                    break
            if not found_pp:
                post_processors.append(TrimTrailingWhitespace())
        return post_processors

    @classmethod
    def _handle_post_processors(
        cls,
        post_processors: typing.Optional[typing.List["nunavut.postprocessors.PostProcessor"]],
        target_language: typing.Optional["nunavut.lang.Language"],
    ) -> typing.Optional[typing.List["nunavut.postprocessors.PostProcessor"]]:
        """
        Used by constructor to process an optional list of post-processors and to augment or create this list
        if needed to support language options.
        """
        if target_language is not None:

            try:
                limit_empty_lines = target_language.get_config_value("limit_empty_lines")
                post_processors = cls.__augment_post_processors_with_ln_limit_empty_lines(
                    post_processors, int(limit_empty_lines)
                )
            except KeyError:
                pass

            if target_language.get_config_value_as_bool("trim_trailing_whitespace"):
                post_processors = cls.__augment_post_processors_with_ln_trim_trailing_whitespace(post_processors)

        return post_processors

    def __init__(
        self,
        namespace: nunavut.Namespace,
        generate_namespace_types: YesNoDefault = YesNoDefault.DEFAULT,
        templates_dir: typing.Optional[typing.Union[pathlib.Path, typing.List[pathlib.Path]]] = None,
        followlinks: bool = False,
        trim_blocks: bool = False,
        lstrip_blocks: bool = False,
        additional_filters: typing.Optional[typing.Dict[str, typing.Callable]] = None,
        additional_tests: typing.Optional[typing.Dict[str, typing.Callable]] = None,
        additional_globals: typing.Optional[typing.Dict[str, typing.Any]] = None,
        post_processors: typing.Optional[typing.List["nunavut.postprocessors.PostProcessor"]] = None,
        builtin_template_path: str = DEFAULT_TEMPLATE_PATH,
    ):

        super().__init__(namespace, generate_namespace_types)

        if templates_dir is not None and not isinstance(templates_dir, list):
            templates_dir = [templates_dir]

        language_context = self._namespace.get_language_context()
        target_language = language_context.get_target_language()

        self._dsdl_template_loader = DSDLTemplateLoader(
            templates_dirs=templates_dir,
            package_name_for_templates=(
                None if target_language is None else target_language.get_templates_package_name()
            ),
            followlinks=followlinks,
            builtin_template_path=builtin_template_path,
        )

        self._post_processors = self._handle_post_processors(post_processors, target_language)

        self._env = CodeGenEnvironment(
            lctx=language_context,
            loader=self._dsdl_template_loader,
            lstrip_blocks=lstrip_blocks,
            trim_blocks=trim_blocks,
            additional_filters=additional_filters,
            additional_tests=additional_tests,
            additional_globals=additional_globals,
        )

    @property
    def dsdl_loader(self) -> DSDLTemplateLoader:
        return self._dsdl_template_loader

    @property
    def language_context(self) -> nunavut.lang.LanguageContext:
        return self._namespace.get_language_context()

    # +-----------------------------------------------------------------------+
    # | PROTECTED
    # +-----------------------------------------------------------------------+
    def _handle_overwrite(self, output_path: pathlib.Path, allow_overwrite: bool) -> None:
        if output_path.exists():
            if allow_overwrite:
                output_path.chmod(output_path.stat().st_mode | 0o220)
            else:
                raise PermissionError("{} exists and allow_overwrite is False.".format(output_path))

    # +-----------------------------------------------------------------------+
    # | AbstractGenerator
    # +-----------------------------------------------------------------------+

    def get_templates(self) -> typing.Iterable[pathlib.Path]:
        """
        Enumerate all templates found in the templates path.
        :data:`~TEMPLATE_SUFFIX` as the suffix for the filename.

        :return: A list of paths to all templates found by this Generator object.
        """
        return self._dsdl_template_loader.get_templates()

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    @staticmethod
    def _filter_and_write_line(
        line_and_lineend: typing.Tuple[str, str],
        output_file: typing.TextIO,
        line_pps: typing.List["nunavut.postprocessors.LinePostProcessor"],
    ) -> None:
        for line_pp in line_pps:
            line_and_lineend = line_pp(line_and_lineend)
            if line_and_lineend is None:
                raise ValueError(
                    "line post processor must return a 2-tuple. To elide a line return a tuple of empty"
                    "strings. None is not a valid value."
                )

        output_file.write(line_and_lineend[0])
        output_file.write(line_and_lineend[1])

    @classmethod
    def _generate_with_line_buffer(
        cls,
        output_file: typing.TextIO,
        template_gen: typing.Generator[str, None, None],
        line_pps: typing.List["nunavut.postprocessors.LinePostProcessor"],
    ) -> None:
        newline_pattern = re.compile(r"\n|\r\n", flags=re.MULTILINE)
        line_buffer = io.StringIO()
        for part in template_gen:
            search_pos = 0  # type: int
            match_obj = newline_pattern.search(part, search_pos)
            while True:
                if search_pos < 0 or search_pos >= len(part):
                    break
                if match_obj is None:
                    line_buffer.write(part[search_pos:])
                    break

                # We have a newline
                line_buffer.write(part[search_pos : match_obj.start()])
                newline_chars = part[match_obj.start() : match_obj.end()]
                line = line_buffer.getvalue()  # type: str
                line_buffer = io.StringIO()
                cls._filter_and_write_line((line, newline_chars), output_file, line_pps)
                search_pos = match_obj.end()
                match_obj = newline_pattern.search(part, search_pos)
        remainder = line_buffer.getvalue()
        if len(remainder) > 0:
            cls._filter_and_write_line((remainder, ""), output_file, line_pps)

    def _generate_code(
        self,
        output_path: pathlib.Path,
        template: Template,
        template_gen: typing.Generator[str, None, None],
        allow_overwrite: bool,
    ) -> None:
        """
        Logic that should run from _generate_type iff is_dryrun is False.
        """

        self._env.now_utc = datetime.datetime.utcnow()

        from ..lang._common import UniqueNameGenerator

        # reset the name generator state for this type
        UniqueNameGenerator.reset()

        # Predetermine the post processor types.
        line_pps = []  # type: typing.List['nunavut.postprocessors.LinePostProcessor']
        file_pps = []  # type: typing.List['nunavut.postprocessors.FilePostProcessor']
        if self._post_processors is not None:
            for pp in self._post_processors:
                if isinstance(pp, nunavut.postprocessors.LinePostProcessor):
                    line_pps.append(pp)
                elif isinstance(pp, nunavut.postprocessors.FilePostProcessor):
                    file_pps.append(pp)
                else:
                    raise ValueError("PostProcessor type {} is unknown.".format(type(pp)))
        logger.debug("Using post-processors: %r %r", line_pps, file_pps)

        self._handle_overwrite(output_path, allow_overwrite)
        output_path.parent.mkdir(parents=True, exist_ok=True)
        with open(str(output_path), "w") as output_file:
            if len(line_pps) > 0:
                # The logic gets much more complex when doing line post-processing.
                self._generate_with_line_buffer(output_file, template_gen, line_pps)
            else:
                for part in template_gen:
                    output_file.write(part)
        for file_pp in file_pps:
            output_path = file_pp(output_path)


# +---------------------------------------------------------------------------+
# | JINJA : DSDLCodeGenerator
# +---------------------------------------------------------------------------+


class DSDLCodeGenerator(CodeGenerator):
    """
    :class:`~CodeGenerator` implementation that generates code for a given set
    of DSDL types.
    """

    # +-----------------------------------------------------------------------+
    # | JINJA : filters
    # +-----------------------------------------------------------------------+

    @staticmethod
    def filter_yamlfy(value: typing.Any) -> str:
        """
        Filter to, optionally, emit a dump of the dsdl input as a yaml document.
        Available as ``yamlfy`` in all template environments.

        Example::

            /*
            {{ T | yamlfy }}
            */

        Result Example (truncated for brevity)::

            /*
            !!python/object:pydsdl.StructureType
            _attributes:
            - !!python/object:pydsdl.Field
            _serializable: !!python/object:pydsdl.UnsignedIntegerType
                _bit_length: 16
                _cast_mode: &id001 !!python/object/apply:pydsdl.CastMode
                - 0
            _name: value
            */

        :param value: The input value to parse as yaml.

        :return: If a yaml parser is available, a pretty dump of the given value as yaml.
                  If a yaml parser is not available then an empty string is returned.
        """
        return str(yaml_dump(value, Dumper=YamlDumper))

    def filter_type_to_template(self, value: typing.Any) -> str:
        """
        Template for type resolution as a filter. Available as ``type_to_template``
        in all template environments.

        Example::

            {%- for attribute in T.attributes %}
                {%* include attribute.data_type | type_to_template %}
                {%- if not loop.last %},{% endif %}
            {%- endfor %}

        :param value: The input value to change into a template include path.

        :return: A path to a template named for the type with :any:`TEMPLATE_SUFFIX`
        """
        result = self.dsdl_loader.type_to_template(type(value))
        if result is None:
            raise RuntimeError("No template found for type {}".format(type(value)))
        return result.name

    def filter_type_to_include_path(self, value: typing.Any, resolve: bool = False) -> str:
        """
        Emits an include path to the output target for a given type.

        Example::

            # include "{{ T.my_type | type_to_include_path }}"

        Result Example:

            # include "foo/bar/my_type.h"

        :param typing.Any value: The type to emit an include for.
        :param bool resolve: If True the path returned will be absolute else the path will
                             be relative to the folder of the root namespace.
        :return: A string path to output file for the type.
        """

        include_path = self.namespace.find_output_path_for_type(value)
        if resolve:
            return include_path.resolve().as_posix()
        else:
            return include_path.relative_to(self.namespace.output_folder.parent).as_posix()

    @staticmethod
    def filter_typename(value: typing.Any) -> str:
        """
        Filters a given token as its type name. Available as ``typename``
        in all template environments.

        This example supposes that ``T.some_value == "some string"``

        Example::

            {{ T.some_value | typename }}

        Result Example::

            str

        :param value: The input value to filter into a type name.

        :return: The ``__name__`` of the python type.
        """
        return type(value).__name__

    @staticmethod
    def filter_alignment_prefix(offset: pydsdl.BitLengthSet) -> str:
        """
        Provides a string prefix based on a given :class:`pydsdl.BitLengthSet`.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

        .. code-block:: python

            # Given
            B = pydsdl.BitLengthSet(32)

            # and
            template = '{{ B | alignment_prefix }}'

            # then ('str' is stropped to 'str_' before the version is suffixed)
            rendered = 'aligned'

        .. invisible-code-block: python

            jinja_filter_tester(DSDLCodeGenerator.filter_alignment_prefix, template, rendered, 'py', B=B)


        .. code-block:: python

            # Given
            B = pydsdl.BitLengthSet(32)
            B += 1

            # and
            template = '{{ B | alignment_prefix }}'

            # then ('str' is stropped to 'str_' before the version is suffixed)
            rendered = 'unaligned'

        .. invisible-code-block: python

            jinja_filter_tester(DSDLCodeGenerator.filter_alignment_prefix, template, rendered, 'py', B=B)


        :param pydsdl.BitLengthSet offset: A bit length set to test for alignment.
        :return: 'aligned' or 'unaligned' based on the state of the ``offset`` argument.
        """
        if isinstance(offset, pydsdl.BitLengthSet):
            return "aligned" if offset.is_aligned_at_byte() else "unaligned"
        else:  # pragma: no cover
            raise TypeError("Expected BitLengthSet, got {}".format(type(offset).__name__))

    @staticmethod
    def filter_bit_length_set(values: typing.Optional[typing.Union[typing.Iterable[int], int]]) -> pydsdl.BitLengthSet:
        """
        Convert an integer or a list of integers into a :class:`pydsdl.BitLengthSet`.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

            assert type(DSDLCodeGenerator.filter_bit_length_set(23)) == pydsdl.BitLengthSet

        """
        return pydsdl.BitLengthSet(values)

    @staticmethod
    def filter_remove_blank_lines(text: str) -> str:
        """
        Remove blank lines from the supplied string.
        Lines that contain only whitespace characters are also considered blank.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            import pydsdl

            assert DSDLCodeGenerator.filter_remove_blank_lines('123\n  \n\n456\n\t\n\v\f\n789') == '123\n456\n789'

        """
        return re.sub(r"\n([ \t\f\v]*\n)+", r"\n", text)

    @staticmethod
    def filter_bits2bytes_ceil(n_bits: int) -> int:
        """
        Implements ``int(ceil(x/8)) | x >= 0``.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(50) == 7
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(8) == 1
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(7) == 1
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(1) == 1
            assert DSDLCodeGenerator.filter_bits2bytes_ceil(0) == 0

        """
        if n_bits < 0:
            raise ValueError("The number of bits cannot be negative")
        return (int(n_bits) + 7) // 8

    # +-----------------------------------------------------------------------+
    # | JINJA : tests
    # +-----------------------------------------------------------------------+

    @staticmethod
    def is_None(value: typing.Any) -> bool:
        """
        Tests if a value is ``None``

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            assert DSDLCodeGenerator.is_None(None) is True
            assert DSDLCodeGenerator.is_None(1) is False

        """
        return value is None

    @staticmethod
    def is_saturated(t: pydsdl.PrimitiveType) -> bool:
        """
        Tests if a type is a saturated type or not.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl
            import pytest

            saturated_mock = MagicMock(spec=pydsdl.PrimitiveType)
            saturated_mock.cast_mode = pydsdl.PrimitiveType.CastMode.SATURATED
            assert DSDLCodeGenerator.is_saturated(saturated_mock) is True

            truncated_mock = MagicMock(spec=pydsdl.PrimitiveType)
            truncated_mock.cast_mode = pydsdl.PrimitiveType.CastMode.TRUNCATED
            assert DSDLCodeGenerator.is_saturated(truncated_mock) is False

            with pytest.raises(TypeError):
                 DSDLCodeGenerator.is_saturated(MagicMock(spec=pydsdl.SerializableType))

        """
        if isinstance(t, pydsdl.PrimitiveType):
            return {
                pydsdl.PrimitiveType.CastMode.SATURATED: True,
                pydsdl.PrimitiveType.CastMode.TRUNCATED: False,
            }[t.cast_mode]
        else:
            raise TypeError("Cast mode is not defined for {}".format(type(t).__name__))

    @staticmethod
    def is_service_request(instance: pydsdl.Any) -> bool:
        """
        Tests if a type is request type of a service type.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl

            service_request_mock = MagicMock(spec=pydsdl.SerializableType)
            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Request'
            assert DSDLCodeGenerator.is_service_request(service_request_mock) is True

            service_request_mock.has_parent_service = False
            assert DSDLCodeGenerator.is_service_request(service_request_mock) is False

            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Response'
            assert DSDLCodeGenerator.is_service_request(service_request_mock) is False

        """
        return instance.has_parent_service and instance.full_name.split(".")[-1] == "Request"  # type: ignore

    @staticmethod
    def is_service_response(instance: pydsdl.Any) -> bool:
        """
        Tests if a type is response type of a service type.

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl

            service_request_mock = MagicMock(spec=pydsdl.SerializableType)
            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Response'
            assert DSDLCodeGenerator.is_service_response(service_request_mock) is True

            service_request_mock.has_parent_service = False
            assert DSDLCodeGenerator.is_service_response(service_request_mock) is False

            service_request_mock.has_parent_service = True
            service_request_mock.full_name = 'foo.bar.Service_1_0.Request'
            assert DSDLCodeGenerator.is_service_response(service_request_mock) is False

        """
        return instance.has_parent_service and instance.full_name.split(".")[-1] == "Response"  # type: ignore

    @staticmethod
    def is_deprecated(instance: pydsdl.Any) -> bool:
        """
        Tests if a type is marked as deprecated

        .. invisible-code-block: python

            from nunavut.jinja import DSDLCodeGenerator
            from unittest.mock import MagicMock
            import pydsdl

            composite_type_mock = MagicMock(spec=pydsdl.CompositeType)
            composite_type_mock.deprecated = True
            assert DSDLCodeGenerator.is_deprecated(composite_type_mock) is True

            array_type_mock = MagicMock(spec=pydsdl.ArrayType)
            array_type_mock.element_type = composite_type_mock
            assert DSDLCodeGenerator.is_deprecated(array_type_mock) is True

            other_type_mock = MagicMock(spec=pydsdl.SerializableType)
            assert DSDLCodeGenerator.is_deprecated(other_type_mock) is False

        """
        if isinstance(instance, pydsdl.CompositeType):
            return instance.deprecated  # type: ignore
        elif isinstance(instance, pydsdl.ArrayType) and isinstance(instance.element_type, pydsdl.CompositeType):
            return instance.element_type.deprecated  # type: ignore
        else:
            return False

    # +-----------------------------------------------------------------------+

    def __init__(self, namespace: nunavut.Namespace, **kwargs: typing.Any):

        super().__init__(namespace, **kwargs)
        for test_name, test in self._create_all_dsdl_tests().items():
            self._env.add_test(test_name, test)
        self._env.add_conventional_methods_to_environment(self)

    # +-----------------------------------------------------------------------+
    # | AbstractGenerator
    # +-----------------------------------------------------------------------+

    def generate_all(self, is_dryrun: bool = False, allow_overwrite: bool = True) -> typing.Iterable[pathlib.Path]:
        generated = []  # type: typing.List[pathlib.Path]
        provider = self.namespace.get_all_types if self.generate_namespace_types else self.namespace.get_all_datatypes
        for (parsed_type, output_path) in provider():
            logger.info("Generating: %s", parsed_type)
            generated.append(self._generate_type(parsed_type, output_path, is_dryrun, allow_overwrite))
        return generated

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+
    @classmethod
    def _create_instance_tests_for_type(cls, root: pydsdl.Any) -> typing.Dict[str, typing.Callable]:
        tests = dict()

        def _field_is_instance(field_or_datatype: pydsdl.Any) -> bool:
            if isinstance(field_or_datatype, pydsdl.Attribute):
                return isinstance(field_or_datatype.data_type, root)
            else:
                return isinstance(field_or_datatype, root)

        tests[root.__name__] = _field_is_instance
        root_name_lower = root.__name__.lower()
        if len(root_name_lower) > 4 and root_name_lower.endswith("type"):
            tests[root_name_lower[:-4]] = _field_is_instance
        elif len(root_name_lower) > 5 and root_name_lower.endswith("field"):
            tests[root_name_lower[:-5]] = _field_is_instance
        else:
            tests[root_name_lower] = _field_is_instance

        for derived in root.__subclasses__():
            tests.update(cls._create_instance_tests_for_type(derived))
        return tests

    @classmethod
    def _create_all_dsdl_tests(cls) -> typing.Mapping[str, typing.Callable]:
        """
        Create a collection of jinja tests for all base dsdl types.

        .. invisible-code-block: python

            import pydsdl
            from unittest.mock import MagicMock
            from nunavut.jinja import DSDLCodeGenerator

            test_set = DSDLCodeGenerator._create_all_dsdl_tests()

            def _do_pydsdl_instance_test_test(pydsdl_obj, test_name):

                if not test_set[test_name](pydsdl_obj):
                    raise AssertionError(test_name)

            def _do_pydsdl_instance_test_tests(pydsdl_type):
                mock_instance = MagicMock(spec=pydsdl_type)
                _do_pydsdl_instance_test_test(mock_instance, pydsdl_type.__name__)
                if pydsdl_type.__name__.endswith('Type'):
                    _do_pydsdl_instance_test_test(mock_instance, pydsdl_type.__name__[:-4].lower())
                if pydsdl_type.__name__.endswith('Field'):
                    _do_pydsdl_instance_test_test(mock_instance, pydsdl_type.__name__[:-5].lower())
                mock_attribute = MagicMock(spec=pydsdl.Attribute)
                mock_attribute.data_type = mock_instance
                _do_pydsdl_instance_test_test(mock_attribute, pydsdl_type.__name__)

            _do_pydsdl_instance_test_tests(pydsdl.SerializableType)
            _do_pydsdl_instance_test_tests(pydsdl.PrimitiveType)
            _do_pydsdl_instance_test_tests(pydsdl.IntegerType)
            _do_pydsdl_instance_test_tests(pydsdl.ServiceType)

        """
        all_tests = dict()
        all_tests.update(cls._create_instance_tests_for_type(pydsdl.SerializableType))
        all_tests.update(cls._create_instance_tests_for_type(pydsdl.Attribute))
        return all_tests

    def _generate_type(
        self, input_type: pydsdl.CompositeType, output_path: pathlib.Path, is_dryrun: bool, allow_overwrite: bool
    ) -> pathlib.Path:
        template_name = self.filter_type_to_template(input_type)
        template = self._env.get_template(template_name)
        template_gen = template.generate(T=input_type)
        if not is_dryrun:
            self._generate_code(output_path, template, template_gen, allow_overwrite)
        return output_path


# +---------------------------------------------------------------------------+
# | JINJA : SupportGenerator
# +---------------------------------------------------------------------------+


class SupportGenerator(CodeGenerator):
    """
    Generates output files by copying them from within the Nunavut package itself
    for non templates but uses jinja to generate headers from templates with the
    language environment provided but no ``T`` (DSDL type) global set.
    This generator always copies files from those returned by the ``file_iterator``
    to locations under :func:`nunavut.Namespace.get_support_output_folder()`
    """

    def __init__(self, namespace: nunavut.Namespace, **kwargs: typing.Any):

        super().__init__(namespace, builtin_template_path="support", **kwargs)

        target_language = self.language_context.get_target_language()

        self._sub_folders = None  # type: typing.Optional[pathlib.Path]
        self._support_enabled = False  # If not enabled then we remove any support files found
        if target_language is not None:
            self._support_enabled = not target_language.omit_serialization_support

            #  Create the sub-folder to copy-to based on the support namespace.
            self._sub_folders = pathlib.Path("")

            for namespace_part in target_language.support_namespace:
                self._sub_folders = self._sub_folders / pathlib.Path(namespace_part)

    # +-----------------------------------------------------------------------+
    # | AbstractGenerator
    # +-----------------------------------------------------------------------+
    def get_templates(self) -> typing.Iterable[pathlib.Path]:
        files = []
        target_language = self.language_context.get_target_language()

        if target_language is not None:
            for resource in target_language.support_files:
                files.append(resource)
        return files

    def generate_all(self, is_dryrun: bool = False, allow_overwrite: bool = True) -> typing.Iterable[pathlib.Path]:
        target_language = self.language_context.get_target_language()
        if self._sub_folders is None or target_language is None:
            logger.info("No target language, therefore, no support headers")
            return []
        else:
            return self._generate_all(target_language, self._sub_folders, is_dryrun, allow_overwrite)

    # +-----------------------------------------------------------------------+
    # | Private
    # +-----------------------------------------------------------------------+

    def _generate_all(
        self, target_language: nunavut.lang.Language, sub_folders: pathlib.Path, is_dryrun: bool, allow_overwrite: bool
    ) -> typing.Iterable[pathlib.Path]:
        target_path = pathlib.Path(self.namespace.get_support_output_folder()) / sub_folders

        line_pps = []  # type: typing.List['nunavut.postprocessors.LinePostProcessor']
        file_pps = []  # type: typing.List['nunavut.postprocessors.FilePostProcessor']
        if self._post_processors is not None:
            for pp in self._post_processors:
                if isinstance(pp, nunavut.postprocessors.LinePostProcessor):
                    line_pps.append(pp)
                elif isinstance(pp, nunavut.postprocessors.FilePostProcessor):
                    file_pps.append(pp)
                else:
                    raise ValueError("PostProcessor type {} is unknown.".format(type(pp)))

        generated = []  # type: typing.List[pathlib.Path]
        for resource in self.get_templates():
            target = (target_path / resource.name).with_suffix(target_language.extension)
            logger.info("Generating support file: %s", target)
            if not self._support_enabled:
                self._remove_header(target, is_dryrun, allow_overwrite)
            elif resource.suffix == TEMPLATE_SUFFIX:
                self._generate_header(resource, target, is_dryrun, allow_overwrite)
                generated.append(target)
            else:
                self._copy_header(resource, target, is_dryrun, allow_overwrite, line_pps, file_pps)
                generated.append(target)
        return generated

    def _remove_header(self, target: pathlib.Path, is_dryrun: bool, allow_overwrite: bool) -> None:
        if not is_dryrun:
            if not allow_overwrite and target.exists():
                raise PermissionError("{} exists. Refusing to remove.".format(str(target)))
            try:
                target.unlink()
            except FileNotFoundError:
                # missing_ok was added in python 3.8 so this try/except statement will
                # go away someday when python 3.7 support is dropped.
                pass

    def _generate_header(
        self, template_path: pathlib.Path, output_path: pathlib.Path, is_dryrun: bool, allow_overwrite: bool
    ) -> pathlib.Path:
        template = self._env.get_template(template_path.name)
        template_gen = template.generate()
        if not is_dryrun:
            self._generate_code(output_path, template, template_gen, allow_overwrite)
        return output_path

    def _copy_header(
        self,
        resource: pathlib.Path,
        target: pathlib.Path,
        is_dryrun: bool,
        allow_overwrite: bool,
        line_pps: typing.List["nunavut.postprocessors.LinePostProcessor"],
        file_pps: typing.List["nunavut.postprocessors.FilePostProcessor"],
    ) -> pathlib.Path:

        if not is_dryrun:
            self._handle_overwrite(target, allow_overwrite)
            target.parent.mkdir(parents=True, exist_ok=True)
            if len(line_pps) == 0:
                shutil.copy(str(resource), str(target))
            else:
                self._copy_header_using_line_pps(resource, target, line_pps)
            for file_pp in file_pps:
                target = file_pp(target)
        return target

    def _copy_header_using_line_pps(
        self,
        resource: pathlib.Path,
        target: pathlib.Path,
        line_pps: typing.List["nunavut.postprocessors.LinePostProcessor"],
    ) -> None:
        with open(str(target), "w") as target_file:
            with open(str(resource), "r") as resource_file:
                for resource_line in resource_file:
                    if len(resource_line) > 1 and resource_line[-2] == "\r":
                        resource_line_tuple = (resource_line[0:-2], "\r\n")
                    else:
                        resource_line_tuple = (resource_line[0:-1], "\n")
                    for line_pp in line_pps:
                        resource_line_tuple = line_pp(resource_line_tuple)
                    target_file.write(resource_line_tuple[0])
                    target_file.write(resource_line_tuple[1])



#
# Copyright 2020 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2020  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""Language-specific support in nunavut.

This package contains modules that provide specific support for generating
source for various languages using templates.
"""
import functools
import pathlib
import re
import typing

import pydsdl

from . import Language


class IncludeGenerator:
    def __init__(self, language: Language, t: pydsdl.CompositeType):
        self._type = t
        self._language = language

    def generate_include_filepart_list(self, output_extension: str, sort: bool) -> typing.List[str]:
        dep_types = self._language.get_dependency_builder(self._type).direct()

        path_list = [
            self.make_path(dt, self._language, output_extension).as_posix() for dt in dep_types.composite_types
        ]

        if not self._language.omit_serialization_support:
            namespace_path = pathlib.Path("")
            for namespace_part in self._language.support_namespace:
                namespace_path = namespace_path / pathlib.Path(namespace_part)
            path_list += [
                (namespace_path / pathlib.Path(p.name).with_suffix(output_extension)).as_posix()
                for p in self._language.support_files
            ]

        prefer_system_includes = self._language.get_config_value_as_bool("prefer_system_includes", False)
        if prefer_system_includes:
            path_list_with_punctuation = ["<{}>".format(p) for p in path_list]
        else:
            path_list_with_punctuation = ['"{}"'.format(p) for p in path_list]

        if sort:
            return sorted(path_list_with_punctuation) + self._language.get_includes(dep_types)
        else:
            return path_list_with_punctuation + self._language.get_includes(dep_types)

    @classmethod
    def make_path(
        cls,
        dt: pydsdl.CompositeType,
        language: typing.Optional[Language] = None,
        output_extension: typing.Optional[str] = None,
    ) -> pathlib.Path:
        """
        Common method for createing a relative path to a datatype source file.

        .. invisible-code-block: python

            import pydsdl
            from nunavut.lang._common import IncludeGenerator
            from nunavut.lang import Language
            from unittest.mock import MagicMock

            config = {
                        'nunavut.lang.c':
                        {
                            'enable_stropping': True
                        }
                    }

            lctx = configurable_language_context_factory(config, 'c')
            lang_c = lctx.get_target_language()

            test_type = MagicMock(spec=pydsdl.CompositeType)
            test_type.parent_service = False
            test_type.attributes = []
            test_type.full_namespace = 'name.space'
            test_type.short_name = 'typename'
            test_type.version = MagicMock()
            test_type.version.major = 1
            test_type.version.minor = 0

        .. code-block:: python

            assert 'name/space/typename_1_0.h' == IncludeGenerator.make_path(test_type, lang_c, '.h').as_posix()
            assert 'name/space/typename_1_0.h' == IncludeGenerator.make_path(test_type, lang_c).as_posix()
            assert 'name/space/typename_1_0' == IncludeGenerator.make_path(test_type).as_posix()

        """
        if language is None:
            short_name = "{short}_{major}_{minor}".format(
                short=dt.short_name, major=dt.version.major, minor=dt.version.minor
            )
        else:
            short_name = language.filter_short_reference_name(dt, id_type="path")

        if output_extension is None:
            output_extension = "" if language is None else language.extension

        ns_path = pathlib.Path(*cls._make_ns_list(language, dt)) / pathlib.Path(short_name).with_suffix(
            output_extension
        )
        return ns_path

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    @classmethod
    def _make_ns_list(cls, language: typing.Optional[Language], dt: pydsdl.SerializableType) -> typing.List[str]:
        if language is not None and language.enable_stropping:
            return [language.filter_id(x, id_type="path") for x in dt.full_namespace.split(".")]
        else:
            return typing.cast(typing.List[str], dt.full_namespace.split("."))


class UniqueNameGenerator:
    """
    Functor used by template filters to obtain a unique name within a given template.
    This should be made available as a private global within each template.
    """

    _singleton = None  # type: typing.Optional['UniqueNameGenerator']

    def __init__(self) -> None:
        self._index_map = {}  # type: typing.Dict[str, typing.Dict[str, int]]

    @classmethod
    def reset(cls) -> None:
        cls._singleton = cls()

    @classmethod
    def get_instance(cls) -> "UniqueNameGenerator":
        if cls._singleton is None:
            raise RuntimeError("No UniqueNameGenerator has been created. Please use reset to create.")
        return cls._singleton

    def __call__(self, key: str, base_token: str, prefix: str, suffix: str) -> str:
        """
        Uses a global index to generate a number unique to a given base_token within a template
        for a given domain (key).
        """
        try:
            keymap = self._index_map[key]
        except KeyError:
            keymap = {}
            self._index_map[key] = keymap

        try:
            next_index = keymap[base_token]
            keymap[base_token] = next_index + 1
        except KeyError:
            next_index = 0
            keymap[base_token] = 1

        return "{prefix}{base_token}{index}{suffix}".format(
            prefix=prefix, base_token=base_token, index=next_index, suffix=suffix
        )


class TokenEncoder:
    """
    One-way transforms from strings of unicode characters into valid identifiers for a given language.

    The behavior of the encoder is entirely driven by language configuration. In the examples below the
    following configuration is used:

    .. code-block:: python
            stropping_prefix = '_pre_'
            stropping_suffix = '_post_'
            encoding_prefix = '_code_'
            reserved_token_patterns_by_type = {
                'var':
                    [
                        '^reserved[A-Za-z]',
                        '^[A-Z]+',
                        '^(__)|(^(_)[A-Z])'
                    ]
            }
            reserved_identifiers = [
                'this_is_reserved',
                '_pre_reservedToken_post_'
            ]
            token_encoding_rules_by_identifier_type = {
                'all':
                    [
                        '(^\\d{1})|([^a-zA-Z0-9_]+)'
                    ]
            }
            whitespace_encoding_char = '_'

    .. invisible-code-block: python

        from nunavut.lang._common import TokenEncoder
        from nunavut.lang import Language

        config = {
                    'nunavut.lang.c':
                    {
                        'stropping_prefix': stropping_prefix,
                        'stropping_suffix': stropping_suffix,
                        'encoding_prefix': encoding_prefix,
                        'reserved_token_patterns_by_type': reserved_token_patterns_by_type,
                        'reserved_identifiers': reserved_identifiers,
                        'token_encoding_rules_by_identifier_type': token_encoding_rules_by_identifier_type,
                        'whitespace_encoding_char': whitespace_encoding_char
                     }
                }

        lctx = configurable_language_context_factory(config, 'c')
        lang_c = lctx.get_target_language()

    .. code-block:: python

        encoder = TokenEncoder(lang_c)
        assert '_code_0061' == encoder.encode_character('a')
        assert 'this_is_not_reserved' == encoder.strop('this_is_not_reserved')
        assert '_pre_this_is_reserved_post_' == encoder.strop('this_is_reserved', 'var')
        assert '_pre_reservedVariableName_post_' == encoder.strop('reservedVariableName', 'var')
        assert '_code_0031CantStartWithNumber' == encoder.strop('1CantStartWithNumber')

        # this shows that encoding happens first then stropping
        assert '_pre_Mem_set_post_' == encoder.strop('Mem set')

        try:
            encoder.strop('reservedToken')
            assert False
        except RuntimeError as e:
            # strop raises RuntimeError if the stropping results in an illegal identifier
            pass

    Languages can also provide additional reserved identifiers in the constructor:

    .. code-block:: python

        encoder = TokenEncoder(lang_c, ['this_is_not_reserved'])

        assert '_pre_this_is_not_reserved_post_' == encoder.strop('this_is_not_reserved')

    .. invisible-code-block: python

        stropping_prefix = '_CODE_'
        encoding_prefix = '_XX_'
        token_encoding_rules_by_identifier_type = {
            'all':
                [
                    '(^\\d{1})|([^a-zA-Z0-9_]+)',
                    '^(__)|(^(_)[A-Z])'
                ]
        }
        config['nunavut.lang.c']['stropping_prefix'] = stropping_prefix
        config['nunavut.lang.c']['encoding_prefix'] = encoding_prefix
        config['nunavut.lang.c']['token_encoding_rules_by_identifier_type'] = token_encoding_rules_by_identifier_type

        lctx = configurable_language_context_factory(config, 'c')
        lang_c = lctx.get_target_language()

        encoder = TokenEncoder(lang_c)

        try:
            encoder.strop('1CantStartWithNumber')
            assert False
        except RuntimeError:
            # strop raises a RuntimeError if the final result is itself reserved.
            pass

    Languages may also provide a handler to be invoked if the automatic stropping fails.

    .. code-block:: python

        def stroppingErrorHandler(encoder: TokenEncoder,
                                 failed_stop: str,
                                 token_type: str,
                                 pending_error: RuntimeError) -> str:
            return 'well_crap'

        encoder = TokenEncoder(lang_c, stropping_failure_handler=stroppingErrorHandler)

        assert 'well_crap' == encoder.strop('reservedToken')

    .. note::
        The token type 'any' is invalid. Any is used to specify patterns that apply to all other token types.
        To strop a token such that it obeys all rules for all token types use the type 'all' instead.

        .. code-block:: python

            encoder = TokenEncoder(lang_c)

            try:
                encoder.strop('foobar', 'all')
                assert False
            except ValueError:
                pass

    """

    def __init__(
        self,
        language: Language,
        additional_reserved_identifiers: typing.Optional[typing.List[str]] = None,
        stropping_failure_handler: typing.Optional[
            typing.Callable[["TokenEncoder", str, str, RuntimeError], str]
        ] = None,
        encoding_failure_handler: typing.Optional[
            typing.Callable[["TokenEncoder", str, str, RuntimeError], str]
        ] = None,
    ) -> None:
        self._reserved_token_patterns_by_type = self._get_map_of_type_to_lists_of_patterns(
            language, "reserved_token_patterns_by_type"
        )
        self._stropping_failure_handler = stropping_failure_handler
        self._encoding_failure_handler = encoding_failure_handler
        self._token_encoding_rules_by_identifier_type = self._get_map_of_type_to_lists_of_patterns(
            language, "token_encoding_rules_by_identifier_type"
        )
        self._reserved_identifiers = language.get_config_value_as_list("reserved_identifiers", default_value=[])
        if additional_reserved_identifiers is not None:
            self._reserved_identifiers = self._reserved_identifiers + additional_reserved_identifiers
        self._stropping_prefix = language.get_config_value("stropping_prefix", "")
        self._stropping_suffix = language.get_config_value("stropping_suffix", "")
        self._encoding_prefix = language.get_config_value("encoding_prefix", "")
        try:
            self._whitespace_encoding_char = language.get_config_value(
                "whitespace_encoding_char"
            )  # type: typing.Optional[str]
        except KeyError:
            self._whitespace_encoding_char = None
        self._collapse_whitespace_when_encoding = language.get_config_value_as_bool("collapse_whitespace_when_encoding")

    def _encoding_filter(self, m: typing.Match) -> str:
        """
        This will encode any illegal characters in an identifier using Python's re.sub function.
        """
        matched_span = m.string[m.start() : m.end()]
        if self._collapse_whitespace_when_encoding and matched_span.isspace():
            if self._whitespace_encoding_char is not None:
                return self._whitespace_encoding_char
            else:
                return self.encode_character(" ")
        else:
            return "".join(map(self.encode_character, matched_span))

    def _matches(
        self, input_string: str, patterns: typing.Union[typing.List[typing.Pattern], typing.List[str]]
    ) -> bool:
        for string_or_pattern in patterns:
            if isinstance(string_or_pattern, str):
                if string_or_pattern == input_string:
                    return True
            elif string_or_pattern.match(input_string):
                return True
        return False

    def _encode(self, token: str, token_type: str, dry_run: bool) -> str:
        encoded = token
        try:
            encoding_rules = self._token_encoding_rules_by_identifier_type[token_type]

            for token_pattern in encoding_rules:
                if not dry_run:
                    encoded = token_pattern.sub(self._encoding_filter, encoded)
                elif token_pattern.match(encoded):
                    raise RuntimeError(
                        'Unstable encoding: using prefix "{}" partially encoded token: "{}"'.format(
                            self._encoding_prefix, encoded
                        )
                    )
        except KeyError:
            pass
        return encoded

    def _strop_by_keyword(self, token: str, token_type: str, dry_run: bool) -> str:
        stropped = token

        if self._matches(stropped, self._reserved_identifiers):
            if not dry_run:
                stropped = self._stropping_prefix + stropped + self._stropping_suffix
            else:
                raise RuntimeError(
                    'input token "{}" of type "{}" yielded an illegal token after '
                    "stropping: {}".format(stropped, token_type, stropped)
                )

        return stropped

    def _strop_by_pattern(self, token: str, token_type: str, dry_run: bool) -> str:

        stropped = token

        reserved_pattern_rules = self._reserved_token_patterns_by_type[token_type]

        if self._matches(stropped, reserved_pattern_rules):
            if not dry_run:
                stropped = self._stropping_prefix + stropped + self._stropping_suffix
            else:
                raise RuntimeError(
                    'input token "{}" of type "{}" yielded an illegal token after '
                    "stropping: {}".format(stropped, token_type, stropped)
                )

        return stropped

    def _do_for_type_and_all(
        self, transform: typing.Callable[[str, str, bool], str], token: str, token_type: str, dry_run: bool
    ) -> str:
        transformed = token

        try:
            transformed = transform(transformed, "all", dry_run)
        except KeyError:
            pass

        if token_type != "all":
            try:
                transformed = transform(transformed, token_type, dry_run)
            except KeyError:
                pass

        return transformed

    # +------------------------------------------------------------------------------------------------------------+
    # | PUBLIC API
    # +------------------------------------------------------------------------------------------------------------+

    def encode_character(self, c: str) -> str:
        if self._whitespace_encoding_char is not None and c.isspace():
            return self._whitespace_encoding_char
        else:
            return "{}{:04X}".format(self._encoding_prefix, ord(c))

    @functools.lru_cache(maxsize=1024)
    def strop(self, token: str, token_type: str = "any") -> str:  # noqa: C901

        token_type_lower = token_type.lower()
        if token_type_lower == "all":
            raise ValueError(
                """Token type 'all' is reserved for patterns that apply to all other types. A single token
                can't be all token types at once but it can be compatible with any type; perhaps you meant 'any'?
            """
            )

        # we encode first.
        encoded = self._do_for_type_and_all(self._encode, token, token_type_lower, False)

        # next we strop the encoded result if it matches a known keyword in the language
        stropped = self._do_for_type_and_all(self._strop_by_keyword, encoded, token_type_lower, False)

        # we then strop if a reserved pattern matches the token
        stropped = self._do_for_type_and_all(self._strop_by_pattern, stropped, token_type_lower, False)

        # and check that the stropping yielded a viable token
        try:
            self._do_for_type_and_all(self._strop_by_pattern, stropped, token_type_lower, True)
        except RuntimeError as pending_error:
            if self._stropping_failure_handler is None:
                raise pending_error
            else:
                stropped = self._stropping_failure_handler(self, stropped, token_type, pending_error)

        # and check that the stropping didn't result in a keyword
        try:
            self._do_for_type_and_all(self._strop_by_keyword, stropped, token_type_lower, True)
        except RuntimeError as pending_error:
            if self._stropping_failure_handler is None:
                raise pending_error
            else:
                stropped = self._stropping_failure_handler(self, stropped, token_type, pending_error)

        # finally, we make sure stropping didn't result in encoding violations
        try:
            self._do_for_type_and_all(self._encode, stropped, token_type_lower, True)
        except RuntimeError as pending_error:
            if self._encoding_failure_handler is None:
                raise pending_error
            else:
                stropped = self._encoding_failure_handler(self, stropped, token_type, pending_error)

        return stropped

    # +----------------------------------------------------------------------------------------------------------------+
    # | Language CONFIGURATION HELPERS
    # +----------------------------------------------------------------------------------------------------------------+
    @classmethod
    def _get_map_of_type_to_lists_of_patterns(
        cls, language: Language, key: str
    ) -> typing.Mapping[str, typing.List[typing.Pattern]]:
        """
        Parses a dictionary value retrieved from Language configuration as a map (with string keys) with each
        value being a list containing pre-compiled, regular expressions.

        Configuration Example:

        .. code-block:: python

            # given this configuration...
            reserved_token_patterns_by_type = {
                    'function':
                        [
                            '^(is|to|str|mem|wcs)[a-z]'
                        ],
                    'typedef':
                        [
                            '^u?int[a-zA-Z_0-9]*_t',
                            '^(atomic_|memory_)[a-z]',
                            '^(cnd_|mtx_|thrd_|tss_)[a-z]'
                        ]
                }

        .. invisible-code-block: python

            from nunavut.lang import Language
            from nunavut.lang._common import TokenEncoder

            config = {
                        'nunavut.lang.cpp':
                        {
                            'reserved_token_patterns_by_type': reserved_token_patterns_by_type
                        }
                    }

            lctx = configurable_language_context_factory(config, 'cpp')
            lang_cpp = lctx.get_target_language()

        .. code-block:: python

            # ... an example use would look like this:
            pattern_map = TokenEncoder._get_map_of_type_to_lists_of_patterns(lang_cpp,
                                                                             'reserved_token_patterns_by_type')
            assert len(pattern_map['function']) == 1
            assert list(pattern_map['function'])[0].match('memset')

        .. invisible-code-block: python

            assert not list(pattern_map['function'])[0].match('foobar')
            assert len(pattern_map['typedef']) == 3

        Note that identifiers are always forced to lower-case:

        .. code-block:: python

            # 'MyIdentiFierTYpE' will be converted to 'myidentifiertype' by this object.
            reserved_token_patterns_by_type = {
                    'MyIdentiFierTYpE':
                        [
                            'foobar'
                        ]
                }

        .. invisible-code-block: python

            config = {
                        'nunavut.lang.cpp':
                        {
                            'reserved_token_patterns_by_type': reserved_token_patterns_by_type
                        }
                    }

            lctx = configurable_language_context_factory(config, 'cpp')
            lang_cpp = lctx.get_target_language()

        .. code-block:: python

            pattern_map = TokenEncoder._get_map_of_type_to_lists_of_patterns(lang_cpp,
                                                                             'reserved_token_patterns_by_type')
            assert len(pattern_map['myidentifiertype']) == 1
            assert list(pattern_map['myidentifiertype'])[0].match('foobar')

        The 'any' key is generated as a union of all keys. If an existing 'any' key is found an exception is raised.
        Any applies to requests to encode or strop a token such that it could be any kind of identifier in a given
        target language. This is different from the special 'all' key which contains patterns that apply to all tokens
        for a target language even if the requested operation is for a specific type.

        .. code-block:: python

            reserved_token_patterns_by_type = {
                    'all':
                        [
                            'all'
                        ],
                    'foo':
                        [
                            'two',
                            'three'
                        ],
                    'bar':
                        [
                            'four',
                            'five'
                        ]
                }

        .. invisible-code-block: python

            config = {
                        'nunavut.lang.cpp':
                        {
                            'reserved_token_patterns_by_type': reserved_token_patterns_by_type
                        }
                    }

            lctx = configurable_language_context_factory(config, 'cpp')
            lang_cpp = lctx.get_target_language()

        .. code-block:: python

            pattern_map = TokenEncoder._get_map_of_type_to_lists_of_patterns(lang_cpp,
                                                                             'reserved_token_patterns_by_type')
            assert len(pattern_map['all']) == 1
            assert len(pattern_map['foo']) == 2
            assert len(pattern_map['bar']) == 2

        .. invisible-code-block: python

            reserved_token_patterns_by_type = {
                    'any':
                        [
                            'foo'
                        ],
                    'all':
                        [
                            'all'
                        ],
                    'foo':
                        [
                            'two',
                            'three'
                        ]
                }
            config['nunavut.lang.cpp']['reserved_token_patterns_by_type'] = reserved_token_patterns_by_type
            lctx = configurable_language_context_factory(config, 'cpp')
            lang_cpp = lctx.get_target_language()

            try:
                TokenEncoder._get_map_of_type_to_lists_of_patterns(lang_cpp,
                                                                   'reserved_token_patterns_by_type')
                assert False
            except RuntimeError:
                pass

        """
        map_of_list_of_patterns = dict()  # type: typing.Dict[str, typing.List[typing.Pattern]]
        map_of_list_of_strings = language.get_config_value_as_dict(key, default_value={})

        any_patterns = []  # type: typing.List[typing.Pattern]
        for mixed_case_type, patterns_for_type in map_of_list_of_strings.items():
            identifier_type = mixed_case_type.lower()
            pattern_list = [re.compile(p) for p in patterns_for_type]
            map_of_list_of_patterns[identifier_type] = pattern_list
            if identifier_type == "any":
                raise RuntimeError(
                    "{}:{} - 'any' key is reserved and cannot be used in configuration.".format(language.name, key)
                )
            any_patterns = any_patterns + pattern_list

        map_of_list_of_patterns["any"] = any_patterns

        return map_of_list_of_patterns



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Objects and utilities for handling DSDL dependencies when generating code for a given type.
"""

import typing

import pydsdl


class Dependencies:
    """
    Data structure that contains a set of composite types and annotations (bool flags)
    which constitute a set of dependencies for a set of DSDL objects.
    """

    def __init__(self) -> None:
        self.composite_types = set()  # type: typing.Set[pydsdl.CompositeType]
        self.uses_integer = False
        self.uses_float = False
        self.uses_variable_length_array = False
        self.uses_array = False
        self.uses_bool = False
        self.uses_primitive_static_array = False
        self.uses_union = False


class DependencyBuilder:
    """
    Given a list of DSDL types this object builds a set of types that the given types use.

    .. invisible-code-block: python

        import pydsdl
        from unittest.mock import MagicMock
        from nunavut.dependencies import DependencyBuilder

        my_dependant_type_l2 = MagicMock(spec=pydsdl.CompositeType)
        my_dependant_type_l2.parent_service = False
        my_dependant_type_l2.attributes = []

        my_dependant_type_l1 = MagicMock(spec=pydsdl.CompositeType)
        my_dependant_type_l1.parent_service = False
        my_dependant_type_l1.attributes = [MagicMock(data_type=my_dependant_type_l2)]

        my_top_level_type = MagicMock(spec=pydsdl.UnionType)
        my_top_level_type.parent_service = False
        my_top_level_type.attributes = [MagicMock(data_type=my_dependant_type_l1)]

        direct_dependencies = DependencyBuilder(my_top_level_type).direct()

        assert len(direct_dependencies.composite_types) == 1
        assert my_dependant_type_l1 in direct_dependencies.composite_types

        transitive_dependencies = DependencyBuilder(my_top_level_type).transitive()

        assert len(transitive_dependencies.composite_types) == 2
        assert my_dependant_type_l1 in transitive_dependencies.composite_types
        assert my_dependant_type_l2 in transitive_dependencies.composite_types
        assert direct_dependencies.uses_integer
        assert direct_dependencies.uses_union

    :param dependant_types: A list of types to build dependencies for.
    :type dependant_types: typing.Iterable[pydsdl.Any]
    """

    def __init__(self, *dependant_types: pydsdl.Any):
        self._dependent_types = dependant_types

    def transitive(self) -> Dependencies:
        """
        Build a set of all transitive dependencies for the dependent types
        set for this builder.
        """
        return self._build_dependency_list(self._dependent_types, True)

    def direct(self) -> Dependencies:
        """
        Build a set of all first-order dependencies for the dependent types
        set for this builder.
        """
        return self._build_dependency_list(self._dependent_types, False)

    # +-----------------------------------------------------------------------+
    # | PRIVATE
    # +-----------------------------------------------------------------------+

    @classmethod
    def _build_dependency_list(
        cls, dependant_types: typing.Iterable[pydsdl.CompositeType], transitive: bool
    ) -> Dependencies:
        results = Dependencies()
        for dependant in dependant_types:
            if isinstance(dependant, pydsdl.UnionType):
                # Unions always require integer for the tag field.
                results.uses_integer = True
                results.uses_union = True
            cls._extract_dependent_types(cls._extract_data_types(dependant), transitive, results)
        return results

    @classmethod
    def _extract_data_types(cls, t: pydsdl.CompositeType) -> typing.List[pydsdl.SerializableType]:
        # Make a list of all attributes defined by this type
        if isinstance(t, pydsdl.ServiceType):
            return [attr.data_type for attr in t.request_type.attributes] + [
                attr.data_type for attr in t.response_type.attributes
            ]
        else:
            return [attr.data_type for attr in t.attributes]

    @classmethod
    def _extract_dependent_types_handle_array_type(
        cls, dependant_type: pydsdl.ArrayType, transitive: bool, inout_dependencies: Dependencies
    ) -> None:
        if isinstance(dependant_type, pydsdl.VariableLengthArrayType):
            inout_dependencies.uses_variable_length_array = True
        else:
            inout_dependencies.uses_array = True
            if isinstance(dependant_type.element_type, pydsdl.PrimitiveType):
                inout_dependencies.uses_primitive_static_array = True

    @classmethod
    def _extract_dependent_types(
        cls, dependant_types: typing.Iterable[pydsdl.Any], transitive: bool, inout_dependencies: Dependencies
    ) -> None:
        for dt in dependant_types:
            if isinstance(dt, pydsdl.CompositeType):
                if dt not in inout_dependencies.composite_types:
                    inout_dependencies.composite_types.add(dt)
                    if transitive:
                        cls._extract_dependent_types(cls._extract_data_types(dt), transitive, inout_dependencies)
            elif isinstance(dt, pydsdl.ArrayType):
                cls._extract_dependent_types_handle_array_type(dt, transitive, inout_dependencies)
                cls._extract_dependent_types([dt.element_type], transitive, inout_dependencies)
            elif isinstance(dt, pydsdl.IntegerType):
                inout_dependencies.uses_integer = True
            elif isinstance(dt, pydsdl.FloatType):
                inout_dependencies.uses_float = True
            elif isinstance(dt, pydsdl.BooleanType):
                inout_dependencies.uses_bool = True



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#

import typing

from nunavut.jinja.jinja2 import Environment, TemplateAssertionError, UndefinedError, nodes
from nunavut.jinja.jinja2.ext import Extension
from nunavut.jinja.jinja2.parser import Parser


class JinjaAssert(Extension):
    """
    Jinja2 extension that allows ``{% assert T.attribute %}`` statements. Templates should
    use these statements where False values would result in malformed source code :

        .. code-block:: python

            template = '''{% assert False %}'''

        .. invisible-code-block: python

            from nunavut.jinja.jinja2.exceptions import TemplateAssertionError
            from nunavut.jinja import CodeGenEnvironment
            from nunavut.jinja.jinja2 import DictLoader
            from nunavut.jinja.extensions import JinjaAssert

            e = CodeGenEnvironment(loader=DictLoader({'test': template}),
                                   extensions=[JinjaAssert])
            try:
                e.get_template('test').render()
                # huh. This should have raised a TemplateAssertionError
                assert False
            except TemplateAssertionError:
                pass

        This extension also support provding an assertion message:

        .. code-block:: python

            template = '''{% assert (1 + 1 == 4) and (2 + 2 == 5), "this was truly false" %}'''

        .. invisible-code-block: python

            e = CodeGenEnvironment(loader=DictLoader({'test': template}),
                                   extensions=[JinjaAssert])
            try:
                e.get_template('test').render()
                # huh. This should have raised a TemplateAssertionError
                assert False
            except TemplateAssertionError as e:
                assert e.message == 'this was truly false'

    """

    tags = set(["assert"])

    def __init__(self, environment: Environment):
        super().__init__(environment)

    def parse(self, parser: Parser) -> nodes.Node:
        """
        See http://jinja.pocoo.org/docs/2.10/extensions/ for help writing
        extensions.
        """

        # This will be the macro name "assert"
        token = next(parser.stream)

        # now we parse a single expression that must evaluate to True
        args = [
            parser.parse_expression(),
            nodes.Const(token.lineno),
            nodes.Const(parser.name),
            nodes.Const(parser.filename),
        ]

        if parser.stream.skip_if("comma"):
            args.append(parser.parse_expression())
        else:
            args.append(nodes.Const("Template assertion failed."))

        return nodes.CallBlock(self.call_method("_do_assert", args), [], [], "").set_lineno(token.lineno)

    def _do_assert(
        self, expression: str, lineno: int, name: str, filename: str, message: str, caller: typing.Callable
    ) -> typing.Any:
        if not expression:
            raise TemplateAssertionError(message, lineno, name, filename)
        return caller()


class UseQuery(Extension):
    """
    Jinja2 extension that allows conditional blocks like ``{% ifuses "std_variant" %}`` or
    ``{% ifnuses "std_variant" %}``. These are defined by the :class:`nunavut.lang.Language` object based on the values
    returned from :meth:`nunavut.lang.Language.get_uses_queries`.

        .. code-block:: python

            template  = ''' {%- ifuses "some_language_key" -%}
                                #include "header 0"
                            {%- elifuses "some_other_language_key" -%}
                                #include "header 1"
                            {%- else -%}
                                #include "header 2"
                            {%- endifuses -%}
                        '''

        .. invisible-code-block: python

            from nunavut.jinja.jinja2.exceptions import TemplateAssertionError
            from nunavut.jinja import CodeGenEnvironment
            from nunavut.jinja.jinja2 import DictLoader
            from nunavut.jinja.extensions import UseQuery
            from nunavut.lang import LanguageLoader
            from nunavut.jinja.jinja2 import UndefinedError
            from unittest.mock import MagicMock

            ln_c = LanguageLoader().load_language('c', True)

            lctx = MagicMock()
            lctx.get_supported_languages = MagicMock(return_value = {'c': ln_c})
            lctx.get_target_language = MagicMock(return_value = ln_c)


            e = CodeGenEnvironment(lctx=lctx,
                                   loader=DictLoader({'test': template}),
                                   extensions=[UseQuery])

            try:
                result = e.get_template('test').render()
                assert False
            except UndefinedError:
                pass

        For "not uses" replace all "uses" tokens with "nuses":

        .. code-block:: python

            template  = ''' {%- ifnuses "some_language_key" -%}
                                #include "header 1"
                            {%- elifnuses "some_other_language_key" -%}
                                #include "header 0"
                            {%- elifuses "yet_another_language_key" -%}
                                #include "header 2"
                            {%- else -%}
                                #include "header 3"
                            {%- endifnuses -%}
                        '''

    """

    tags = set(["ifuses", "ifnuses"])

    def __init__(self, environment: Environment):
        super().__init__(environment)

    def parse(self, parser: Parser) -> nodes.Node:
        """
        See http://jinja.pocoo.org/docs/2.10/extensions/ for help writing
        extensions.
        """

        if parser.stream.current.test("name:ifnuses"):
            negate = True
            ifname = "name:ifnuses"
        else:
            negate = False
            ifname = "name:ifuses"

        node = result = nodes.If(lineno=parser.stream.expect(ifname).lineno)
        while 1:
            args = [
                parser.parse_expression(),
                nodes.Const(parser.stream.current.lineno),
                nodes.Const(parser.name),
                nodes.Const(parser.filename),
            ]
            test_name = "_use_query" if not negate else "_use_nquery"
            node.test = self.call_method(test_name, args)
            node.body = parser.parse_statements(
                ("name:elifuses", "name:elifnuses", "name:else", "name:endifuses", "name:endifnuses")
            )
            node.elif_ = []
            node.else_ = []
            token = next(parser.stream)
            if token.test("name:elifuses"):
                negate = False
                node = nodes.If(lineno=parser.stream.current.lineno)
                result.elif_.append(node)
                continue
            elif token.test("name:elifnuses"):
                negate = True
                node = nodes.If(lineno=parser.stream.current.lineno)
                result.elif_.append(node)
                continue
            elif token.test("name:else"):
                result.else_ = parser.parse_statements(
                    (
                        "name:endifuses",
                        "name:endifnuses",
                    ),
                    drop_needle=True,
                )
            break
        return result

    def _use_query_common(self, uses_query_name: str, lineno: int, name: str, filename: str) -> bool:

        target_language = self.environment.target_language

        if target_language is None:
            raise TemplateAssertionError(
                "ifuses directive cannot be used in a language without a target language.", lineno, name, filename
            )
        if uses_query_name is None:
            raise TemplateAssertionError("Unknown uses_query_name found.", lineno, name, filename)

        try:
            uses_query = typing.cast(
                typing.Callable[..., bool], getattr(self.environment.target_language_uses_queries, uses_query_name)
            )
        except AttributeError:
            raise UndefinedError(
                'use query "{}" for language "{}" is not defined '
                "(line={}, name={}, filename={})".format(uses_query_name, target_language.name, lineno, name, filename)
            )

        return uses_query()

    def _use_nquery(self, uses_query_name: str, lineno: int, name: str, filename: str) -> bool:
        return not self._use_query_common(uses_query_name, lineno, name, filename)

    def _use_query(self, uses_query_name: str, lineno: int, name: str, filename: str) -> bool:
        return self._use_query_common(uses_query_name, lineno, name, filename)



#
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""

.. autodata:: __version__
"""

__version__ = "1.8.2"  #: The version number used in the release of nunavut to pypi.


__license__ = "MIT"



import sys
import types
from ast import literal_eval
from itertools import islice, chain
from nunavut.jinja.jinja2 import nodes
from nunavut.jinja.jinja2._compat import text_type
from nunavut.jinja.jinja2.compiler import CodeGenerator, has_safe_repr
from nunavut.jinja.jinja2.environment import Environment, Template
from nunavut.jinja.jinja2.utils import concat, escape


def native_concat(nodes):
    """Return a native Python type from the list of compiled nodes. If the
    result is a single node, its value is returned. Otherwise, the nodes are
    concatenated as strings. If the result can be parsed with
    :func:`ast.literal_eval`, the parsed value is returned. Otherwise, the
    string is returned.
    """
    head = list(islice(nodes, 2))

    if not head:
        return None

    if len(head) == 1:
        out = head[0]
    else:
        if isinstance(nodes, types.GeneratorType):
            nodes = chain(head, nodes)
        out = u''.join([text_type(v) for v in nodes])

    try:
        return literal_eval(out)
    except (ValueError, SyntaxError, MemoryError):
        return out


class NativeCodeGenerator(CodeGenerator):
    """A code generator which avoids injecting ``to_string()`` calls around the
    internal code Jinja uses to render templates.
    """

    def visit_Output(self, node, frame):
        """Same as :meth:`CodeGenerator.visit_Output`, but do not call
        ``to_string`` on output nodes in generated code.
        """
        if self.has_known_extends and frame.require_output_check:
            return

        finalize = self.environment.finalize
        finalize_context = getattr(finalize, 'contextfunction', False)
        finalize_eval = getattr(finalize, 'evalcontextfunction', False)
        finalize_env = getattr(finalize, 'environmentfunction', False)

        if finalize is not None:
            if finalize_context or finalize_eval:
                const_finalize = None
            elif finalize_env:
                def const_finalize(x):
                    return finalize(self.environment, x)
            else:
                const_finalize = finalize
        else:
            def const_finalize(x):
                return x

        # If we are inside a frame that requires output checking, we do so.
        outdent_later = False

        if frame.require_output_check:
            self.writeline('if parent_template is None:')
            self.indent()
            outdent_later = True

        # Try to evaluate as many chunks as possible into a static string at
        # compile time.
        body = []

        for child in node.nodes:
            try:
                if const_finalize is None:
                    raise nodes.Impossible()

                const = child.as_const(frame.eval_ctx)
                if not has_safe_repr(const):
                    raise nodes.Impossible()
            except nodes.Impossible:
                body.append(child)
                continue

            # the frame can't be volatile here, because otherwise the as_const
            # function would raise an Impossible exception at that point
            try:
                if frame.eval_ctx.autoescape:
                    if hasattr(const, '__html__'):
                        const = const.__html__()
                    else:
                        const = escape(const)

                const = const_finalize(const)
            except Exception:
                # if something goes wrong here we evaluate the node at runtime
                # for easier debugging
                body.append(child)
                continue

            if body and isinstance(body[-1], list):
                body[-1].append(const)
            else:
                body.append([const])

        # if we have less than 3 nodes or a buffer we yield or extend/append
        if len(body) < 3 or frame.buffer is not None:
            if frame.buffer is not None:
                # for one item we append, for more we extend
                if len(body) == 1:
                    self.writeline('%s.append(' % frame.buffer)
                else:
                    self.writeline('%s.extend((' % frame.buffer)

                self.indent()

            for item in body:
                if isinstance(item, list):
                    val = repr(native_concat(item))

                    if frame.buffer is None:
                        self.writeline('yield ' + val)
                    else:
                        self.writeline(val + ',')
                else:
                    if frame.buffer is None:
                        self.writeline('yield ', item)
                    else:
                        self.newline(item)

                    close = 0

                    if finalize is not None:
                        self.write('environment.finalize(')

                        if finalize_context:
                            self.write('context, ')

                        close += 1

                    self.visit(item, frame)

                    if close > 0:
                        self.write(')' * close)

                    if frame.buffer is not None:
                        self.write(',')

            if frame.buffer is not None:
                # close the open parentheses
                self.outdent()
                self.writeline(len(body) == 1 and ')' or '))')

        # otherwise we create a format string as this is faster in that case
        else:
            format = []
            arguments = []

            for item in body:
                if isinstance(item, list):
                    format.append(native_concat(item).replace('%', '%%'))
                else:
                    format.append('%s')
                    arguments.append(item)

            self.writeline('yield ')
            self.write(repr(concat(format)) + ' % (')
            self.indent()

            for argument in arguments:
                self.newline(argument)
                close = 0

                if finalize is not None:
                    self.write('environment.finalize(')

                    if finalize_context:
                        self.write('context, ')
                    elif finalize_eval:
                        self.write('context.eval_ctx, ')
                    elif finalize_env:
                        self.write('environment, ')

                    close += 1

                self.visit(argument, frame)
                self.write(')' * close + ', ')

            self.outdent()
            self.writeline(')')

        if outdent_later:
            self.outdent()


class NativeTemplate(Template):
    def render(self, *args, **kwargs):
        """Render the template to produce a native Python type. If the result
        is a single node, its value is returned. Otherwise, the nodes are
        concatenated as strings. If the result can be parsed with
        :func:`ast.literal_eval`, the parsed value is returned. Otherwise, the
        string is returned.
        """
        vars = dict(*args, **kwargs)

        try:
            return native_concat(self.root_render_func(self.new_context(vars)))
        except Exception:
            exc_info = sys.exc_info()

        return self.environment.handle_exception(exc_info, True)


class NativeEnvironment(Environment):
    """An environment that renders templates to native Python types."""

    code_generator_class = NativeCodeGenerator
    template_class = NativeTemplate



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Filters for generating docs. All filters in this
    module will be available in the template's global namespace as ``ln.html``.
"""

import html
import logging
import re
import typing

import nunavut
import pydsdl

from ...jinja.jinja2 import TemplateAssertionError
from ...templates import template_volatile_filter
from .._common import UniqueNameGenerator

logger = logging.getLogger(__name__)


def filter_extent(instance: pydsdl.Any) -> int:
    try:
        return instance.extent or 0
    except TypeError as e:
        raise TemplateAssertionError(e)


def filter_max_bit_length(instance: pydsdl.Any) -> int:
    try:
        return instance.bit_length_set.max or 0
    except TypeError as e:
        raise TemplateAssertionError(e)


def filter_tag_id(instance: pydsdl.Any) -> str:
    if isinstance(instance, pydsdl.ArrayType):
        return "{}_array".format(str(instance.element_type).replace(".", "_").replace(" ", "_"))
    else:
        return "{}_{}_{}".format(
            instance.full_name.replace(".", "_"),
            instance.version[0],
            instance.version[1],
        )


def filter_url_from_type(instance: pydsdl.Any) -> str:
    root_ns = instance.root_namespace
    tag_id = "{}_{}_{}".format(instance.full_name.replace(".", "_"), instance.version[0], instance.version[1])
    return "../{}/#{}".format(root_ns, tag_id)


@template_volatile_filter
def filter_make_unique(_: typing.Any, base_token: str) -> str:
    """
    Filter that takes a base token and forms a name that is very
    likely to be unique within the template the filter is invoked.

    .. IMPORTANT::

        The exact tokens generated may change between major or minor versions
        of this library. The only guarantee provided is that the tokens
        will be stable for the same version of this library given the same
        input.

        Also note that name uniqueness is only likely within a given template.
        Between templates there is no guarantee of uniqueness and,
        since this library does not lex generated source, there is no guarantee
        that the generated name does not conflict with a name generated by
        another means.

    .. invisible-code-block: python

        from nunavut.lang.html import filter_make_unique
        from nunavut.lang._common import UniqueNameGenerator

    .. code-block:: python

        # Given
        template  = '{{ "foo" | make_unique }},{{ "Foo" | make_unique }},'
        template += '{{ "fOO" | make_unique }}'

        # then
        rendered = 'foo0,foo1,fOO0'

    .. invisible-code-block: python

        UniqueNameGenerator.reset()
        jinja_filter_tester(filter_make_unique, template, rendered, 'html')

    .. code-block:: python

        # Given
        template = '{{ "coffee > tea" | make_unique }}'

        # then
        rendered = 'coffee &gt; tea0'

    .. invisible-code-block: python

        UniqueNameGenerator.reset()
        jinja_filter_tester(filter_make_unique, template, rendered, 'html')


    :param str base_token: A token to include in the base name.
    :return: A name that is likely to be unique within the file generated by the current
             template.
    """
    if len(base_token) > 0:
        adj_base_token = base_token[0:1].lower() + base_token[1:]
    else:
        adj_base_token = base_token

    escaped_base_token = html.escape(adj_base_token)

    return UniqueNameGenerator.get_instance()("html", escaped_base_token, "", "")


def filter_namespace_doc(ns: nunavut.Namespace) -> str:
    result = ""
    for t, _ in ns.get_nested_types():
        if t.short_name == "_":
            result = t.doc
            break
    return result


def filter_display_type(instance: pydsdl.Any) -> str:
    # TODO: this whole thing needs to be in the template.
    if isinstance(instance, pydsdl.FixedLengthArrayType):
        capacity = '<span style="color: green">[{}]</span>'.format(instance.capacity)
        return filter_display_type(instance.element_type) + capacity
    elif isinstance(instance, pydsdl.VariableLengthArrayType):
        capacity = '<span style="color: green">[<={}]</span>'.format(instance.capacity)
        return filter_display_type(instance.element_type) + capacity
    elif isinstance(instance, pydsdl.PaddingField):
        return '<span style="color: gray">{}</span>'.format(instance)
    elif isinstance(instance, pydsdl.Field):
        return "{} {}".format(filter_display_type(instance.data_type), instance.name)
    elif isinstance(instance, pydsdl.Constant):
        name = '<span style="color: darkmagenta">{}</span>'.format(instance.name)
        value = '<span style="color: darkcyan">{}</span>'.format(instance.value)
        return "{} {} = {}".format(filter_display_type(instance.data_type), name, value)
    elif isinstance(instance, pydsdl.PrimitiveType):
        if instance.cast_mode == instance.cast_mode.SATURATED:
            is_saturated = '<span style="color: gray">saturated</span> '
        else:
            is_saturated = '<span style="color: orange">truncated</span> '
        type_name = '<span style="color: green">{}</span>'.format(str(instance).split()[-1])
        return "{}{}".format(is_saturated, type_name)
    else:
        return str(instance)


def _natural_sort(instance: typing.List[pydsdl.Any], key: typing.Callable = lambda s: s) -> typing.List[pydsdl.Any]:
    def natural_sort_key(s: str, _nsre: typing.Pattern = re.compile("([0-9]+)")) -> typing.List[pydsdl.Any]:
        _key = key(s)

        return [int(text) if text.isdigit() else text.lower() for text in _nsre.split(_key)]

    return sorted(instance, key=natural_sort_key)


def filter_natural_sort_namespace(instance: typing.List[pydsdl.Any]) -> typing.List[pydsdl.Any]:
    """
    Namespaces come in plain lists; sort by name only.
    """
    return _natural_sort(instance, key=lambda s: s.full_name)


def filter_natural_sort_type(instance: pydsdl.Any) -> typing.List[pydsdl.Any]:
    """
    Types come in tuples (type, path). Sort by type name.
    """
    return _natural_sort(instance, key=lambda s: s[0].full_name)



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
A small collection of common utilities.

.. note::

    Please don't use this as a dumping ground for things that belong in a dedicated package. Python being such a
    full-featured language, there should be very few truly generic utilities in Nunavut.

"""
import enum
import logging
import pathlib
from typing import Generator, cast

import importlib_resources

_logger = logging.getLogger(__name__)


@enum.unique
class YesNoDefault(enum.Enum):
    """
    Trinary type for decisions that allow a default behavior to be requested that can
    be different based on other contexts. For example:

    .. invisible-code-block: python

        from datetime import datetime
        from nunavut._utilities import YesNoDefault

    .. code-block:: python

        def should_we_order_pizza(answer: YesNoDefault) -> bool:
            if answer == YesNoDefault.YES or (
               answer == YesNoDefault.DEFAULT and
               datetime.today().isoweekday() == 5):
                # if yes or if we are taking the default action which is to
                # order pizza on Friday, and today is Friday, then we order pizza
                return True
            else:
                return False

    .. invisible-code-block: python

        assert should_we_order_pizza(YesNoDefault.YES)
        assert not should_we_order_pizza(YesNoDefault.NO)

    """

    @classmethod
    def test_truth(cls, ynd_value: "YesNoDefault", default_value: bool) -> bool:
        """
        Helper method to test a YesNoDefault value and return a default boolean value.

        .. invisible-code-block: python

            from nunavut._utilities import YesNoDefault

        .. code-block:: python

            '''
                let "is YES" be Y
                let "is DEFAULT" be D where:
                    if Y then not D and if D then not Y
                    and "is NO" is Y = D = 0
                let "is default_value true" be d

                Y | D | d | Y or (D and d)
                1   *   *    1
                0   1   0    0
                0   1   1    1
                0   0   *    0
            '''

            assert YesNoDefault.test_truth(YesNoDefault.YES, False)
            assert not YesNoDefault.test_truth(YesNoDefault.DEFAULT, False)
            assert YesNoDefault.test_truth(YesNoDefault.DEFAULT, True)
            assert not YesNoDefault.test_truth(YesNoDefault.NO, True)

        """
        if ynd_value == cls.DEFAULT:
            return default_value
        else:
            return ynd_value == cls.YES

    NO = 0
    YES = 1
    DEFAULT = 2


def iter_package_resources(pkg_name: str, *suffix_filters: str) -> Generator[pathlib.Path, None, None]:
    """
    >>> from nunavut._utilities import iter_package_resources
    >>> rs = [x for x in iter_package_resources("nunavut.lang", ".py") if x.name == "__init__.py"]
    >>> len(rs)
    1
    >>> rs[0].name
    '__init__.py'

    """
    for resource in importlib_resources.files(pkg_name).iterdir():
        if any(suffix == resource.suffix for suffix in suffix_filters):  # type: ignore
            yield cast(pathlib.Path, resource)



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
Module containing types and utilities for building generator objects.
Generators abstract the code generation technology used to transform
pydsdl AST into source code.
"""

import abc
import pathlib
import typing

import nunavut
from nunavut._utilities import YesNoDefault


class AbstractGenerator(metaclass=abc.ABCMeta):
    """
    Abstract base class for classes that generate source file output
    from a given pydsdl parser result.

    :param nunavut.Namespace namespace:  The top-level namespace to
        generates types at and from.
    :param YesNoDefault generate_namespace_types:  Set to YES
        to force generation files for namespaces and NO to suppress.
        DEFAULT will generate namespace files based on the language
        preference.
    """

    def __init__(
        self,
        namespace: nunavut.Namespace,
        generate_namespace_types: YesNoDefault = YesNoDefault.DEFAULT,
    ):
        self._namespace = namespace
        if generate_namespace_types == YesNoDefault.YES:
            self._generate_namespace_types = True
        elif generate_namespace_types == YesNoDefault.NO:
            self._generate_namespace_types = False
        else:
            target_language = self._namespace.get_language_context().get_target_language()
            if target_language is not None and target_language.has_standard_namespace_files:
                self._generate_namespace_types = True
            else:
                self._generate_namespace_types = False

    @property
    def namespace(self) -> nunavut.Namespace:
        """
        The root :class:`nunavut.Namespace` for this generator.
        """
        return self._namespace

    @property
    def generate_namespace_types(self) -> bool:
        """
        If true then the generator is set to emit files for :class:`nunavut.Namespace`
        in addition to the pydsdl datatypes. If false then only files for pydsdl datatypes
        will be generated.
        """
        return self._generate_namespace_types

    @abc.abstractmethod
    def get_templates(self) -> typing.Iterable[pathlib.Path]:
        """
        Enumerate all templates found in the templates path.
        :return: A list of paths to all templates found by this Generator object.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def generate_all(self, is_dryrun: bool = False, allow_overwrite: bool = True) -> typing.Iterable[pathlib.Path]:
        """
        Generates all output for a given :class:`nunavut.Namespace` and using
        the templates found by this object.

        :param bool is_dryrun: If True then no output files will actually be
                               written but all other operations will be performed.
        :param bool allow_overwrite: If True then the generator will attempt to overwrite any existing files
                                it encounters. If False then the generator will raise an error if the
                                output file exists and the generation is not a dry-run.
        :return: 0 for success. Non-zero for errors.
        :raises: PermissionError if :attr:`allow_overwrite` is False and the file exists.
        """
        raise NotImplementedError()


def create_generators(
    namespace: nunavut.Namespace, **kwargs: typing.Any
) -> typing.Tuple["AbstractGenerator", "AbstractGenerator"]:
    """
    Create the two generators used by Nunavut; a code-generator and a support-library generator.

    :param nunavut.Namespace namespace: The namespace to generate code within.
    :param kwargs: A list of arguments that are forwarded to the generator constructors.
    :return: Tuple with the first item being the code-generator and the second the support-library
        generator.
    """
    from nunavut.jinja import DSDLCodeGenerator, SupportGenerator

    return (DSDLCodeGenerator(namespace, **kwargs), SupportGenerator(namespace, **kwargs))



#
# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2019  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
"""
    Filters for generating C++. All filters in this
    module will be available in the template's global namespace as ``cpp``.
"""

import functools
import fractions
import io
import re
import textwrap
import typing

import pydsdl

from ...templates import template_language_filter, template_language_list_filter, template_language_test
from .. import Dependencies
from .. import Language as BaseLanguage
from .._common import IncludeGenerator, TokenEncoder, UniqueNameGenerator
from ..._utilities import YesNoDefault
from ..c import _CFit
from ..c import filter_literal as c_filter_literal

DEFAULT_ARRAY_TYPE = "std::array<{TYPE},{MAX_SIZE}>"


class Language(BaseLanguage):
    """
    Concrete, C++-specific :class:`nunavut.lang.Language` object.
    """

    CPP_STD_EXTRACT_NUMBER_PATTERN = re.compile(r"(?:gnu|c)\+\+(\d(?:\w))")

    @staticmethod
    def _handle_stropping_or_encoding_failure(
        encoder: TokenEncoder, stropped: str, token_type: str, pending_error: RuntimeError
    ) -> str:
        """
        If the generic stropping fails we take one last look to see if there is something c++-specific we can do.
        """
        # Note that this is imprecise because C++ does not allow identifiers to start with an underscore if they are in
        # the global namespace, however, we don't have an AST for the C++ we're generating so there's no way to know if
        # the given token is global. Since this library is about generating code from DSDL and no DSDL identifier should
        # be in the global namespace (except for the top-level namespace for the datatypes) we assume that the token is
        # not.
        m = re.match(r"^_+([A-Z]?)", stropped)
        if m:
            # Resolve the conflict between C's global identifier rules and our desire to use
            # '_' as a stropping prefix:
            return "_{}{}".format(m.group(1).lower(), stropped[m.end() :])

        # we couldn't help after all. raise the pending error.
        raise pending_error

    @functools.lru_cache(maxsize=None)
    def _get_token_encoder(self) -> TokenEncoder:
        """
        Caching getter to ensure we don't have to recompile TokenEncoders for each filter invocation.
        """
        return TokenEncoder(
            self,
            stropping_failure_handler=self._handle_stropping_or_encoding_failure,
            encoding_failure_handler=self._handle_stropping_or_encoding_failure,
        )

    def _standard_version(self) -> int:
        """
        .. invisible-code-block: python

           from nunavut.lang import LanguageLoader

           # test c++17
           language = LanguageLoader().load_language('cpp', True, {'std': 'c++17'})
           assert language._standard_version() == 17

           # test c++14
           language = LanguageLoader().load_language('cpp', True, {'std': 'c++14'})
           assert language._standard_version() == 14

           # test gnu++20
           language = LanguageLoader().load_language('cpp', True, {'std': 'gnu++20'})
           assert language._standard_version() == 20
        """
        std = str(self.get_option("std", ""))

        match = self.CPP_STD_EXTRACT_NUMBER_PATTERN.match(std)

        if match is not None and len(match.groups()) >= 1:
            return int(match.group(1))
        else:
            return 0

    def _has_variant(self) -> bool:
        """
        .. invisible-code-block: python

           from nunavut.lang import LanguageLoader

           # test c++17
           language = LanguageLoader().load_language('cpp', True, {'std': 'c++17'})
           assert language._has_variant()

           # test c++14
           language = LanguageLoader().load_language('cpp', True, {'std': 'c++14'})
           assert not language._has_variant()

           # test gnu++20
           language = LanguageLoader().load_language('cpp', True, {'std': 'gnu++20'})
           assert language._has_variant()
        """
        return self._standard_version() >= 17

    def get_includes(self, dep_types: Dependencies) -> typing.List[str]:
        std_includes = []  # type: typing.List[str]
        if self.get_config_value_as_bool("use_standard_types"):
            if dep_types.uses_integer:
                std_includes.append("cstdint")
            if dep_types.uses_array:
                std_includes.append("array")
            if dep_types.uses_variable_length_array:
                std_includes.append("vector")
            if dep_types.uses_union and self._has_variant():
                std_includes.append("variant")
        return ["<{}>".format(include) for include in sorted(std_includes)]

    def filter_id(self, instance: typing.Any, id_type: str = "any") -> str:
        raw_name = self.default_filter_id_for_target(instance)

        return self._get_token_encoder().strop(raw_name, id_type)


@template_language_test(__name__)
def uses_std_variant(language: Language) -> bool:
    """
    Uses query for std variant.

    If the language options contain an ``std`` entry for C++ and the specified standard includes the
    ``std::variant`` type added to the language at C++17 then this value is true. The logic included
    in this filter can be stated as "options has key std and the value for options.std evaluates to
    C++ version 17 or greater" but the implementation is able to parse out actual compiler flags like
    ``gnu++20`` and is aware of any overrides to suppress use of the standard variant type even if
    available.

    Example:

        .. code-block:: python

            template = '''
                {%- ifuses "std_variant" -%}
                    #include <variant>
                {%- else -%}
                    #include "user_variant.h"
                {%- endifuses -%}
            '''

        .. invisible-code-block: python

            # test c++17
            config_overrides = {'nunavut.lang.cpp': {'options': {'std': 'c++17' }}}
            lctx = configurable_language_context_factory(config_overrides, 'cpp')
            jinja_filter_tester(None, template, '#include <variant>', lctx)

            # test c++14
            config_overrides = {'nunavut.lang.cpp': {'options': {'std': 'c++14' }}}
            lctx = configurable_language_context_factory(config_overrides, 'cpp')
            jinja_filter_tester(None, template, '#include "user_variant.h"', lctx)

    """
    return language._has_variant()


@template_language_filter(__name__)
def filter_constant_value(language: Language, constant: pydsdl.Constant) -> str:
    """
    Renders the specified value of the specified type as a literal.
    """
    return c_filter_literal(language, constant.value.native_value, constant.data_type)


@template_language_filter(__name__)
def filter_literal(
    language: Language,
    value: typing.Union[fractions.Fraction, bool, int],
    ty: pydsdl.Any,
    cast_format: typing.Optional[str] = None,
) -> str:
    """
    Renders the specified value of the specified type as a literal.
    """
    return c_filter_literal(language, value, ty, cast_format)


def filter_to_standard_bit_length(t: pydsdl.PrimitiveType) -> int:
    """
    Returns the nearest standard bit length of a type as an int.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_standard_bit_length
        import pydsdl

    .. code-block: python
        # Given
        I = pydsdl.UnsignedIntegerType(7, pydsdl.PrimitiveType.CastMode.TRUNCATED)

        # and
        template = '{{ I | to_standard_bit_length }}'

        # then
        rendered = '8'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_standard_bit_length, template, rendered, 'c', I=I)

    """
    return int(_CFit.get_best_fit(t.bit_length).value)


@template_language_filter(__name__)
def filter_id(language: Language, instance: typing.Any, id_type: str = "any") -> str:
    """
    Filter that produces a valid C and/or C++ identifier for a given object. The encoding may not
    be reversible.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_id


    .. code-block:: python

        # Given
        I = 'I like c++'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'I_like_czX002BzX002B'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)


    .. code-block:: python

        # Given
        I = 'if'

        # and
        template = '{{ I | id }}'

        # then
        rendered = '_if'


    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)


    .. code-block:: python

        # Given
        I = 'I   really like \t coffee'

        # and
        template = '{{ I | id }}'

        # then
        rendered = 'I_really_like_coffee'

    .. invisible-code-block: python

        jinja_filter_tester(filter_id, template, rendered, 'cpp', I=I)

    :param any instance:        Any object or data that either has a name property or can be converted
                                to a string.
    :return: A token that is a valid identifier for C and C++, is not a reserved keyword, and is transformed
              in a deterministic manner based on the provided instance.
    """
    return language.filter_id(instance, id_type)


@template_language_filter(__name__)
def filter_type(language: Language, obj: typing.Any) -> str:
    """
    Tries to convert a Python object into a c++ typename.

    Will raise a ValueError if the object provided does not (yet) have an available conversion in this function.

    Currently supported types are string:

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_type

    .. code-block:: python

         # given
        template = '{{ "Any" | type }}'

        # then
        rendered = "const char* const"

    .. invisible-code-block: python

        jinja_filter_tester(filter_type, template, rendered, 'cpp')

    int:

    .. code-block:: python

         # given
        template = '{{ 123 | type }}'

        # then
        rendered = 'long long'

    .. invisible-code-block: python

        jinja_filter_tester(filter_type, template, rendered, 'cpp')

    and bool:

    .. code-block:: python

         # given
        template = '{{ True | type }}'

        # then
        rendered = 'bool'

    .. invisible-code-block: python

        jinja_filter_tester(filter_type, template, rendered, 'cpp')

    """

    if isinstance(obj, bool):
        return "bool"
    if isinstance(obj, int):
        return "long long"
    if isinstance(obj, str):
        return "const char* const"

    return filter_type_from_primitive(language, obj)


@template_language_filter(__name__)
def filter_open_namespace(
    language: Language, full_namespace: str, bracket_on_next_line: bool = True, linesep: str = "\n"
) -> str:
    """
    Emits c++ opening namespace syntax parsed from a pydsdl "full_namespace",
    dot-separated  value.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_open_namespace
        T = lambda : None;
        setattr(T, 'full_namespace)', '')

    .. code-block:: python

        # Given
        T.full_namespace = 'uavcan.foo'

        # and
        template = '{{ T.full_namespace | open_namespace }}'

        # then
        rendered = '''namespace uavcan
        {
        namespace foo
        {'''

    .. invisible-code-block: python

        # stropping doesn't change our example here.

        lctx = configurable_language_context_factory({'nunavut.lang.cpp': {'enable_stropping': False}}, 'cpp')
        jinja_filter_tester(filter_open_namespace, template, rendered, lctx, T=T)

        lctx = configurable_language_context_factory({'nunavut.lang.cpp': {'enable_stropping': True}}, 'cpp')
        jinja_filter_tester(filter_open_namespace, template, rendered, lctx, T=T)

    :param str full_namespace: A dot-separated namespace string.
    :param bool bracket_on_next_line: If True (the default) then the opening
        brackets are placed on a newline after the namespace keyword.
    :param str linesep: The line-separator to use when emitting new lines.
                        By default this is ``\\n``.

    :return: C++ namespace declarations with opening brackets.
    """

    with io.StringIO() as content:
        first = True
        for name in full_namespace.split("."):
            if first:
                first = False
            else:
                content.write(linesep)
            content.write("namespace ")
            if language.enable_stropping:
                content.write(language.filter_id(name))
            else:
                content.write(name)
            if bracket_on_next_line:
                content.write(linesep)
            else:
                content.write(" ")
            content.write("{")
        return content.getvalue()


@template_language_filter(__name__)
def filter_close_namespace(
    language: Language, full_namespace: str, omit_comments: bool = False, linesep: str = "\n"
) -> str:
    """
    Emits c++ closing namespace syntax parsed from a pydsdl "full_namespace",
    dot-separated  value.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_close_namespace
        T = lambda : None;
        setattr(T, 'full_namespace)', '')

    .. code-block:: python

        # Given
        T.full_namespace = 'uavcan.foo'

        # and
        template = '{{ T.full_namespace | close_namespace }}'

        # then
        rendered = '''} // namespace foo
        } // namespace uavcan'''

    .. invisible-code-block: python

        jinja_filter_tester(filter_close_namespace, template, rendered, 'cpp', T=T)


    :param str full_namespace: A dot-separated namespace string.
    :param bool omit_comments: If True then the comments following the closing
                               bracket are omitted.
    :param str linesep: The line-separator to use when emitting new lines.
                        By default this is ``\\n``

    :return: C++ namespace declarations with opening brackets.
    """
    with io.StringIO() as content:
        first = True
        for name in reversed(full_namespace.split(".")):
            if first:
                first = False
            else:
                content.write(linesep)

            content.write("}")
            if not omit_comments:
                content.write(" // namespace ")
                if language.enable_stropping:
                    content.write(language.filter_id(name))
                else:
                    content.write(name)
        return content.getvalue()


@template_language_filter(__name__)
def filter_full_reference_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string that is the full namespace, typename, major, and minor version for a given composite type.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_full_reference_name
        from unittest.mock import MagicMock
        import pydsdl

        my_obj = MagicMock()
        my_obj.has_parent_service = False
        my_obj.version = MagicMock()

    .. code-block:: python

        # Given a type with illegal characters for C++
        my_obj.full_name = 'any.int.2Foo'
        my_obj.version.major = 1
        my_obj.version.minor = 2

        # and
        template = '{{ my_obj | full_reference_name }}'

        # then, with stropping enabled
        rendered = 'any::_int::_2Foo_1_2'

    .. invisible-code-block: python

        my_obj.short_name = my_obj.full_name.split('.')[-1]
        my_obj.full_namespace = '.'.join(my_obj.full_name.split('.')[:-1])

        jinja_filter_tester(filter_full_reference_name, template, rendered, 'cpp', my_obj=my_obj)

    :param pydsdl.CompositeType t: The DSDL type to get the fully-resolved reference name for.
    """
    ns_parts = t.full_namespace.split(".")
    if language.enable_stropping:
        ns = list(map(functools.partial(filter_id, language), ns_parts))
    else:
        ns = ns_parts

    full_path = ns + [language.filter_short_reference_name(t)]
    return "::".join(full_path)


@template_language_filter(__name__)
def filter_full_macro_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string usable as part of a macro name that is the full namespace, typename, major, and minor version
    for a given composite type.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_full_macro_name
        from unittest.mock import MagicMock
        import pydsdl

        my_obj = MagicMock()
        my_obj.has_parent_service = False
        my_obj.version = MagicMock()

    .. code-block:: python

        # Given a type with illegal characters for C++
        my_obj.full_name = 'any.int.2Foo'
        my_obj.version.major = 1
        my_obj.version.minor = 2

        # and
        template = '{{ my_obj | full_macro_name }}'

        # then, with stropping enabled
        rendered = 'any__int__2Foo_1_2'

    .. invisible-code-block: python

        my_obj.short_name = my_obj.full_name.split('.')[-1]
        my_obj.full_namespace = '.'.join(my_obj.full_name.split('.')[:-1])

        jinja_filter_tester(filter_full_macro_name, template, rendered, 'cpp', my_obj=my_obj)

    :param pydsdl.CompositeType t: The DSDL type to get the fully-resolved reference name for.
    """
    ns_parts = t.full_namespace.split(".")
    if language.enable_stropping:
        ns = list(map(functools.partial(filter_id, language), ns_parts))
    else:
        ns = ns_parts

    full_path = ns + [language.filter_short_reference_name(t)]
    return "_".join(full_path)


@template_language_filter(__name__)
def filter_short_reference_name(language: Language, t: pydsdl.CompositeType) -> str:
    """
    Provides a string that is a shorted version of the full reference name. This type is unique only within its
    namespace.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_short_reference_name
        from unittest.mock import MagicMock
        import pydsdl

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None
        my_type.has_parent_service = False

    .. code-block:: python

        # Given a type with illegal C++ characters
        my_type.short_name = '2Foo'
        my_type.version.major = 1
        my_type.version.minor = 2

        # and
        template = '{{ my_type | short_reference_name }}'

        # then, with stropping enabled
        rendered = '_2Foo_1_2'

    .. invisible-code-block: python

        jinja_filter_tester(filter_short_reference_name, template, rendered, 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None
        my_type.has_parent_service = False

    .. code-block:: python

        # Given a type with legal C++ characters
        my_type.short_name = 'Struct_'
        my_type.version.major = 0
        my_type.version.minor = 1

        # and
        template = '{{ my_type | short_reference_name }}'

        # then, with stropping enabled
        rendered = 'Struct__0_1'

    .. invisible-code-block: python

        jinja_filter_tester(filter_short_reference_name, template, rendered, 'cpp', my_type=my_type)

    .. invisible-code-block: python
        my_service_type = MagicMock(spec=pydsdl.ServiceType)
        my_service_type.version = MagicMock()
        my_service_type.parent_service = None
        my_service_type.has_parent_service = False
        my_service_type.request_type = MagicMock(spec=pydsdl.StructureType)
        my_service_type.request_type.has_parent_service = True
        my_service_type.request_type.short_name = "Request"
        my_service_type.request_type.version = my_service_type.version
        my_service_type.request_type.parent_service = my_service_type
        my_service_type.response_type = MagicMock(spec=pydsdl.StructureType)
        my_service_type.response_type.has_parent_service = True
        my_service_type.response_type.short_name = "Response"
        my_service_type.response_type.version = my_service_type.version
        my_service_type.response_type.parent_service = my_service_type

    .. code-block:: python

        # Given a service type
        my_service_type.short_name = 'Struct_'
        my_service_type.version.major = 0
        my_service_type.version.minor = 1

        # and
        template = '''
        {{ my_service_type | short_reference_name }}
        {{ my_service_type.request_type | short_reference_name }}
        {{ my_service_type.response_type | short_reference_name }}
        '''

        # then, with stropping enabled
        rendered = '''
        Struct_
        Request_0_1
        Response_0_1
        '''

    .. invisible-code-block: python
        jinja_filter_tester(filter_short_reference_name, template, rendered, 'cpp', my_service_type=my_service_type)

    :param pydsdl.CompositeType t: The DSDL type to get the reference name for.
    """
    if isinstance(t, pydsdl.ServiceType):
        if YesNoDefault.test_truth(YesNoDefault.DEFAULT, language.enable_stropping):
            return language.filter_id(t.short_name)
        else:
            return str(t.short_name)
    return language.filter_short_reference_name(t)


@template_language_list_filter(__name__)
def filter_includes(language: Language, t: pydsdl.CompositeType, sort: bool = True) -> typing.List[str]:
    """
    Returns a list of all include paths for a given type.

    :param pydsdl.CompositeType t: The type to scan for dependencies.
    :param bool sort: If true the returned list will be sorted.
    :return: a list of include headers needed for a given type.
    """
    return IncludeGenerator(language, t).generate_include_filepart_list(language.extension, sort)


@template_language_filter(__name__)
def filter_destructor_name(language: Language, instance: pydsdl.Any) -> str:
    """
    Returns a token that is the local destructor name. For example:

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_destructor_name
        from unittest.mock import MagicMock
        import pydsdl

        my_type = MagicMock(spec=pydsdl.FixedLengthArrayType)
        my_type.element_type = MagicMock(spec=pydsdl.UnsignedIntegerType)
        my_type.element_type.bit_length = 8
        my_type.element_type.cast_mode = pydsdl.PrimitiveType.CastMode.SATURATED


    .. code-block:: python

        # Given a pydsdl.FixedLengthArrayType "my_type":
        my_type.short_name = 'Foo'
        my_type.capacity = 2

        # and
        template = 'ptr->{{ my_type | destructor_name }}'

        # then
        rendered = 'ptr->~array<std::uint8_t,2>'

    .. invisible-code-block: python

        jinja_filter_tester(filter_destructor_name, template, rendered, 'cpp', my_type=my_type)


    :param pydsdl.CompositeType t: The type to generate a destructor template for.
    :return: A destructor name token.
    """
    declaration = filter_declaration(language, instance)
    declaration_parts = declaration.split("<")
    declaration_parts[0] = declaration_parts[0].split("::")[-1]
    return "~" + "<".join(declaration_parts)


@template_language_filter(__name__)
def filter_declaration(language: Language, instance: pydsdl.Any) -> str:
    """
    Emit a declaration statement for the given instance.
    """
    if isinstance(instance, pydsdl.PrimitiveType) or isinstance(instance, pydsdl.VoidType):
        return filter_type_from_primitive(language, instance)
    elif isinstance(instance, pydsdl.VariableLengthArrayType):
        variable_array_type = language.get_option("variable_array_type")

        if not isinstance(variable_array_type, str):
            raise RuntimeError("variable_array_type language option was missing or invalid.")
        return variable_array_type.format(
            TYPE=filter_declaration(language, instance.element_type), MAX_SIZE=instance.capacity
        )
    elif isinstance(instance, pydsdl.ArrayType):
        return DEFAULT_ARRAY_TYPE.format(
            TYPE=filter_declaration(language, instance.element_type), MAX_SIZE=instance.capacity
        )
    else:
        return filter_full_reference_name(language, instance)


@template_language_filter(__name__)
def filter_definition_begin(language: Language, instance: pydsdl.CompositeType) -> str:
    """
    Emit the start of a definition statement for a composite type.

     .. invisible-code-block: python

        from nunavut.lang.cpp import filter_definition_begin
        from unittest.mock import MagicMock
        import pytest
        import pydsdl

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.parent_service = None
        my_type.has_parent_service = False

        with pytest.raises(ValueError):
            jinja_filter_tester(filter_definition_begin,
                                '{{ my_type | definition_begin }}',
                                '',
                                'cpp',
                                my_type=MagicMock())

    .. code-block:: python

        # Given a pydsdl.CompositeType "my_type":
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        # and
        template = '{{ my_type | definition_begin }}'

        # then
        rendered = 'struct Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, template, rendered, 'cpp', my_type=my_type)

        my_union_type = MagicMock(spec=pydsdl.UnionType)
        my_union_type.version = MagicMock()
        my_union_type.parent_service = None
        my_union_type.has_parent_service = False

    .. code-block:: python

        # Also, given a pydsdl.UnionType "my_union_type":
        my_union_type.short_name = 'Foo'
        my_union_type.version.major = 1
        my_union_type.version.minor = 0

        # and
        union_template = '{{ my_union_type | definition_begin }}'

        # then
        rendered = 'struct Foo_1_0'

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, union_template, rendered, 'cpp', my_union_type=my_union_type)

        my_service_type = MagicMock(spec=pydsdl.ServiceType)
        my_service_type.version = MagicMock()
        my_service_type.parent_service = None
        my_service_type.has_parent_service = False
        my_service_type.request_type = MagicMock(spec=pydsdl.StructureType)
        my_service_type.request_type.short_name = "Request"
        my_service_type.request_type.version = my_service_type.version
        my_service_type.request_type.parent_service = my_service_type
        my_service_type.request_type.has_parent_service = True
        my_service_type.response_type = MagicMock(spec=pydsdl.StructureType)
        my_service_type.response_type.short_name = "Response"
        my_service_type.response_type.version = my_service_type.version
        my_service_type.response_type.parent_service = my_service_type
        my_service_type.response_type.has_parent_service = True

    .. code-block:: python

        # Finally, given a pydsdl.ServiceType "my_service_type":
        my_service_type.short_name = 'Foo'
        my_service_type.version.major = 1
        my_service_type.version.minor = 0

        # and
        template = '''
        {{ my_service_type | definition_begin }};
        {{ my_service_type.request_type | definition_begin }};
        {{ my_service_type.response_type | definition_begin }};
        '''

        # then
        rendered = '''
        namespace Foo;
        struct Request_1_0;
        struct Response_1_0;
        '''

    .. invisible-code-block: python

        jinja_filter_tester(filter_definition_begin, template, rendered, 'cpp', my_service_type=my_service_type)

    """
    short_name = filter_short_reference_name(language, instance)
    if (
        isinstance(instance, pydsdl.DelimitedType)
        or isinstance(instance, pydsdl.StructureType)
        or isinstance(instance, pydsdl.UnionType)
    ):
        return "struct {}".format(short_name)
    elif isinstance(instance, pydsdl.ServiceType):
        return "namespace {}".format(short_name)
    else:
        raise ValueError("{} types cannot be redefined.".format(type(instance).__name__))


@template_language_filter(__name__)
def filter_definition_end(language: Language, instance: pydsdl.CompositeType) -> str:
    """
    Emit the end of a definition statement for a composite type.

     .. invisible-code-block: python

        from nunavut.lang.cpp import filter_definition_end
        from unittest.mock import MagicMock
        import pytest
        import pydsdl


        with pytest.raises(ValueError):
            jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', '', 'cpp', my_type=MagicMock())

        my_type = MagicMock(spec=pydsdl.StructureType)
        my_type.version = MagicMock()
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', ';', 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.UnionType)
        my_type.version = MagicMock()
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end, '{{ my_type | definition_end }}', ';', 'cpp', my_type=my_type)

        my_type = MagicMock(spec=pydsdl.ServiceType)
        my_type.version = MagicMock()
        my_type.parent_service = None
        my_type.short_name = 'Foo'
        my_type.version.major = 1
        my_type.version.minor = 0

        jinja_filter_tester(filter_definition_end,
                            '{{ my_type | definition_end }}',
                            ' // namespace Foo_1_0',
                            'cpp',
                            my_type=my_type)

    """
    if (
        isinstance(instance, pydsdl.DelimitedType)
        or isinstance(instance, pydsdl.StructureType)
        or isinstance(instance, pydsdl.UnionType)
    ):
        return ";"
    elif isinstance(instance, pydsdl.ServiceType):
        return " // namespace {}".format(language.filter_short_reference_name(instance))
    else:
        raise ValueError("{} types cannot be redefined.".format(type(instance).__name__))


@template_language_filter(__name__)
def filter_type_from_primitive(language: Language, value: pydsdl.PrimitiveType) -> str:
    """
    Filter to transform a pydsdl :class:`~pydsdl.PrimitiveType` into
    a valid C++ type.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_type_from_primitive
        import pydsdl


    .. code-block:: python

        # Given
        template = '{{ unsigned_int_32_type | type_from_primitive }}'

        # then
        rendered = 'std::uint32_t'

    .. invisible-code-block: python

        test_type = pydsdl.UnsignedIntegerType(32, pydsdl.PrimitiveType.CastMode.TRUNCATED)
        jinja_filter_tester(filter_type_from_primitive,
                            template,
                            rendered,
                            'cpp',
                            unsigned_int_32_type=test_type)

    Also note that this is sensitive to the ``use_standard_types`` configuration in the language properties:

    .. code-block:: python

        # rendered will be different if use_standard_types is False
        rendered = 'unsigned long'

    .. invisible-code-block: python

        lctx = configurable_language_context_factory({'nunavut.lang.cpp': {'use_standard_types': False}}, 'cpp')
        jinja_filter_tester(filter_type_from_primitive,
                            template,
                            rendered,
                            lctx,
                            unsigned_int_32_type=test_type)

    :param str value: The dsdl primitive to transform.

    :return: A valid C++ type name.

    :raises TemplateRuntimeError: If the primitive cannot be represented as a standard C++ type.
    """
    return _CFit.get_best_fit(value.bit_length).to_c_type(value, language, "std::")


def filter_to_namespace_qualifier(namespace_list: typing.List[str]) -> str:
    """
    Converts a list of namespace names into a qualifier string. For example:

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_namespace_qualifier
        import pydsdl


    .. code-block:: python

        my_namespace = ['foo', 'bar']
        template = '{{ my_namespace | to_namespace_qualifier }}myType()'
        expected = 'foo::bar::myType()'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_namespace_qualifier, template, expected, 'cpp', my_namespace=my_namespace)

    This filter gracefully handles empty namespace lists:

    .. code-block:: python

        my_namespace = []
        template = '{{ my_namespace | to_namespace_qualifier }}myType()'
        expected = 'myType()'

    .. invisible-code-block: python

        jinja_filter_tester(filter_to_namespace_qualifier, template, expected, 'cpp', my_namespace=my_namespace)

    """
    if namespace_list is None or len(namespace_list) == 0:
        return ""
    else:
        return "::".join(namespace_list) + "::"


def filter_to_template_unique_name(base_token: str) -> str:
    """
    Filter that takes a base token and forms a name that is very
    likely to be unique within the template the filter is invoked. This
    name is also very likely to be a valid C++ identifier.

    .. IMPORTANT::

        The exact tokens generated may change between major or minor versions
        of this library. The only guarantee provided is that the tokens
        will be stable for the same version of this library given the same
        input.

        Also note that name uniqueness is only likely within a given template.
        Between templates there is no guarantee of uniqueness and,
        since this library does not lex generated source, there is no guarantee
        that the generated name does not conflict with a name generated by
        another means.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_to_template_unique_name
        from nunavut.lang._common import UniqueNameGenerator

    .. code-block:: python

        # Given
        template  = '{{ "foo" | to_template_unique_name }},{{ "Foo" | to_template_unique_name }},'
        template += '{{ "fOO" | to_template_unique_name }}'

        # then
        rendered = '_foo0_,_foo1_,_fOO0_'

    .. invisible-code-block: python

        UniqueNameGenerator.reset()
        jinja_filter_tester(filter_to_template_unique_name, template, rendered, 'cpp')

    .. code-block:: python

        # Given
        template = '{{ "i like coffee" | to_template_unique_name }}'

        # then
        rendered = '_i like coffee0_'

    .. invisible-code-block: python

        UniqueNameGenerator.reset()
        jinja_filter_tester(filter_to_template_unique_name, template, rendered, 'cpp')


    :param str base_token: A token to include in the base name.
    :return: A name that is likely to be valid C++ identifier and is likely to
        be unique within the file generated by the current template.
    """
    if len(base_token) > 0:
        adj_base_token = base_token[0:1].lower() + base_token[1:]
    else:
        adj_base_token = base_token

    return UniqueNameGenerator.get_instance()("cpp", adj_base_token, "_", "_")


def filter_as_boolean_value(value: bool) -> str:
    """
    Filter a boolean expression to produce a valid C++ "true" or "false" token.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_as_boolean_value

    .. code-block:: python

        assert "true" == filter_as_boolean_value(True)
        assert "false" == filter_as_boolean_value(False)

    """
    return "true" if value else "false"


@template_language_filter(__name__)
def filter_indent_if_not(language: Language, text: str, depth: int = 1) -> str:
    r"""
    Emit indent characters as configured for the language but only as needed. This
    is different from the built-in indent filter in that it may add or remove spaces based on the
    existing indent.

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_indent_if_not

    .. code-block:: python

        # Given a string with an existing indent of 4 spaces...
        template  = '{{ "    int a = 1;" | indent_if_not }}'

        # then if cpp.indent == 4 we expect no change.
        rendered = '    int a = 1;'

    .. invisible-code-block: python

        assert_language_config_value('cpp', 'indent', '4', 'this test expected an indent of 4')
        jinja_filter_tester(filter_indent_if_not, template, rendered, 'cpp')

    .. code-block:: python

        # If the indent is only 3 spaces...
        template  = '{{ "   int a = 1;" | indent_if_not }}'

        # then if cpp.indent == 4 we expect 4 spaces (i.e. not 7 spaces)
        rendered = '    int a = 1;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent_if_not, template, rendered, 'cpp')

    .. code-block:: python

        # We can also specify multiple indents...
        template  = '{{ "int a = 1;" | indent_if_not(2) }}'

        rendered = '        int a = 1;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent_if_not, template, rendered, 'cpp')

    .. code-block:: python

        # ...or no indent
        template  = '{{ "    int a = 1;" | indent_if_not(0) }}'

        rendered = 'int a = 1;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent_if_not, template, rendered, 'cpp')

    .. code-block:: python

        # Finally, note that blank lines are not indented.
        template  = '''
            {%- set block_text -%}
                int a = 1;
                {# empty line #}
                int b = 2;
            {%- endset -%}{{ block_text | indent_if_not(1) }}'''

        rendered  = '    int a = 1;'
        rendered += '\n'
        rendered += '\n'  # Nothing but spaces so this is stripped
        rendered += '    int b = 2;'

    .. invisible-code-block: python

        jinja_filter_tester(filter_indent_if_not, template, rendered, 'cpp')

    :param text: The text to indent.
    :param depth: The number of indents. For example, if depth is 2 and the indent for this language is
        4 spaces then the text will be indented by 8 spaces.

    """
    configured_indent = int(language.get_config_value("indent"))
    lines = text.splitlines(keepends=True)
    result = ""
    for i in range(0, len(lines)):
        line = lines[i].lstrip()
        if len(line) == 0:
            # don't indent blank lines
            result += "\n"
        else:
            result += (" " * (depth * configured_indent)) + line

    return result


def filter_minimum_required_capacity_bits(t: pydsdl.SerializableType) -> int:
    """
    Returns the minimum number of bits required to store the deserialized value of a
    pydsdl :class:`~pydsdl.SerializableType`. This capacity may be too small for some
    instances of the value (e.g. variable length arrays).

    .. invisible-code-block: python

        from nunavut.lang.cpp import filter_minimum_required_capacity_bits
        import pydsdl
        import pytest

    .. code-block:: python

        # Given
        template = '{{ unsigned_int_32_type | minimum_required_capacity_bits }}'

        # then
        rendered = '32'

    .. invisible-code-block: python

        uint32_truncated_type = pydsdl.UnsignedIntegerType(32, pydsdl.PrimitiveType.CastMode.TRUNCATED)
        jinja_filter_tester(filter_minimum_required_capacity_bits,
                            template,
                            rendered,
                            'cpp',
                            unsigned_int_32_type=uint32_truncated_type)

        # Cover the other cases here. We don't need to pain the user with these gory
        # details so we'll keep them in the hidden block.
        assert filter_minimum_required_capacity_bits(pydsdl.VoidType(1)) == 1
        assert filter_minimum_required_capacity_bits(pydsdl.FixedLengthArrayType(uint32_truncated_type, 2)) == 64
        assert filter_minimum_required_capacity_bits(pydsdl.VariableLengthArrayType(uint32_truncated_type, 2)) == 8

        field_one = pydsdl.Field(pydsdl.UnsignedIntegerType(32, pydsdl.PrimitiveType.CastMode.TRUNCATED), 'one')
        field_two = pydsdl.Field(pydsdl.UnsignedIntegerType(64, pydsdl.PrimitiveType.CastMode.TRUNCATED), 'two')

        test_struct = pydsdl.StructureType(
            name='uavcan.foo',
            version=pydsdl.Version(0, 1),
            attributes=[field_one, field_two],
            deprecated=False,
            fixed_port_id=None,
            source_file_path='',
            has_parent_service = False
        )

        test_union = pydsdl.UnionType(
            name='uavcan.foo',
            version=pydsdl.Version(0, 1),
            attributes=[field_one, field_two],
            deprecated=False,
            fixed_port_id=None,
            source_file_path='',
            has_parent_service = False
        )

        assert filter_minimum_required_capacity_bits(test_struct) == 96
        assert filter_minimum_required_capacity_bits(test_union) == 32 + 8

    :param pydsdl.SerializableType t: The dsdl type.

    :return: The minimum, required bits needed to store some values of the given type.
    """
    return typing.cast(int, min(t.bit_length_set))


@functools.lru_cache(3)
def _make_textwrap(width: int, initial_indent: str, subseqent_indent: str) -> textwrap.TextWrapper:
    return textwrap.TextWrapper(
        width=width,
        initial_indent=initial_indent,
        subsequent_indent=subseqent_indent,
        break_on_hyphens=True,
        break_long_words=False,
        replace_whitespace=False,
    )


def _make_block_comment(text: str, prefix: str, comment: str, suffix: str, indent: int, line_length: int) -> str:
    doc_lines = text.splitlines()  # type: typing.List[str]
    indented_comment = "{}{}".format(" " * indent, comment)

    commented_doc_lines = []  # type: typing.List[str]

    if len(doc_lines) > 0:
        if len(prefix) > 0:
            commented_doc_lines.append(prefix)
        else:
            commented_doc_lines.extend(
                _make_textwrap(width=line_length, initial_indent=comment, subseqent_indent=indented_comment).wrap(
                    doc_lines.pop(0)
                )
            )

    tw = _make_textwrap(width=line_length, initial_indent=indented_comment, subseqent_indent=indented_comment)

    for docline in doc_lines:
        commented_doc_lines.extend(tw.wrap(docline))

    if len(suffix) > 0 and len(commented_doc_lines) > 0:
        commented_doc_lines.append("{}{}".format(" " * indent, suffix))

    return "\n".join(commented_doc_lines)


@template_language_filter(__name__)
def filter_block_comment(language: Language, text: str, style: str, indent: int = 0, line_length: int = 100) -> str:
    """
    Reformats text as a block comment using Python's :meth:`textwrap.TextWrapper.wrap` function.

    :param text: The text to emit as a block comment.
    :param style: Dictates the style of comments (see return documentation for valid style names).
    :param indent: The number of spaces to indent the comments by (tab indent is not supported. Sorry).
    :param line_length: The soft maximum width to wrap text at. Some violations may occur where long words are used.

    :returns str: A comment block. Comment styles supported are:

        .. invisible-code-block: python

            from nunavut.lang.cpp import filter_block_comment

            # Initial, private, verification:
            text = '''This is a list:
             1. one
             2. two
             3. three'''

            template = '''
                {{ text | block_comment('cpp-doxygen', 4, 50) }}
                void some_method();
            '''

            rendered = '''
                ///
                /// This is a list:
                ///  1. one
                ///  2. two
                ///  3. three
                ///
                void some_method();
            '''

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

            # handle empty case
            text = ''

            template = '''
                {{ text | block_comment('cpp-doxygen', 4, 50) }}
                void some_method();
            '''

            rendered = '''
                {}
                void some_method();
            '''.format('')

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

            template = '''
                {{ text | block_comment('c', 4, 50) }}
                void some_method();
            '''

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

            # Cover ValueError clause
            template = "{{ text | block_comment('not a style', 4, 24) }}"

            try:
                jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)
                assert False
            except ValueError:
                pass

        **javadoc**

        .. code-block:: python

            # Given a type with the following docstring
            text = 'This is a bunch of documentation.'

            # and
            template = '''
                {{ text | block_comment('javadoc', 4, 24) }}
                void some_method();
            '''

            # the output will be
            rendered = '''
                /**
                 * This is a bunch
                 * of documentation.
                 */
                void some_method();
            '''

        .. invisible-code-block: python

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

        **cpp-doxygen**

        .. code-block:: python

            # that same template using the cpp style of doxygen...
            template = '''
                {{ text | block_comment('cpp-doxygen', 4, 24) }}
                void some_method();
            '''

            # ...will be
            rendered = '''
                ///
                /// This is a bunch
                /// of
                /// documentation.
                ///
                void some_method();
            '''

        .. invisible-code-block: python

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

        **cpp**

        .. code-block:: python

            # also supported is cpp style...
            template = '''
                {{ text | block_comment('cpp', 4, 24) }}
                void some_method();
            '''

            rendered = '''
                // This is a bunch of
                // documentation.
                void some_method();
            '''

        .. invisible-code-block: python

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

        **c**

        .. code-block:: python

            # c style...
            template = '''
                {{ text | block_comment('c', 4, 24) }}
                void some_method();
            '''

            rendered = '''
                /*
                 * This is a bunch
                 * of documentation.
                 */
                void some_method();
            '''

        .. invisible-code-block: python

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

        **qt**

        .. code-block:: python

            # and Qt style...
            template = '''
                {{ text | block_comment('qt', 4, 24) }}
                void some_method();
            '''

            rendered = '''
                /*!
                 * This is a bunch
                 * of documentation.
                 */
                void some_method();
            '''

        .. invisible-code-block: python

            jinja_filter_tester(filter_block_comment, template, rendered, 'cpp', text=text)

            from nunavut.lang import LanguageContext

            comment_configs = LanguageContext('cpp').get_target_language().get_config_value_as_dict('comment_styles')

            if len(comment_configs) != 5:
                raise RuntimeError('A comment style was added but not documented here. Please document it/them.')

    """

    config_styles = language.get_config_value_as_dict(
        "comment_styles"
    )  # type: typing.Mapping[str, typing.Mapping[str, str]]

    try:
        config_style = config_styles[style.lower()]
    except KeyError:
        raise ValueError(
            "{} is not a supported comment style. Supported is c, cpp, cpp-doxygen, and javadoc".format(style)
        )

    return _make_block_comment(
        text=text,
        prefix=config_style["prefix"],
        comment=config_style["comment"],
        suffix=config_style["suffix"],
        indent=indent,
        line_length=line_length,
    )



#
# Copyright 2021 Amazon.com, Inc. or its affiliates. All Rights Reserved.
# Copyright (C) 2018-2021  UAVCAN Development Team  <uavcan.org>
# This software is distributed under the terms of the MIT License.
#
import collections
import importlib
import logging
import pathlib
import typing

import pydsdl

from ..lang._config import VersionReader
from .jinja2 import BaseLoader, Environment, FileSystemLoader, PackageLoader, TemplateNotFound

logger = logging.getLogger(__name__)


TEMPLATE_SUFFIX = ".j2"  #: The suffix expected for Jinja templates.

DEFAULT_TEMPLATE_PATH = "templates"


# +--------------------------------------------------------------------------------------------------------------------+
# | LOADERS : DSDLTemplateLoader
# +--------------------------------------------------------------------------------------------------------------------+


class DSDLTemplateLoader(BaseLoader):
    """
    Nunavut's DSDL template loader is similar to a choice loader with a file-system loader
    first and a package loader as a fallback. The major difference is a DFS is performed
    on the type hierarchy of the type a template is being loaded for. So, for example,
    if no ``StructureType.j2`` template is found then this loader will look for a ``CompositeType.j2``
    and so on.

    :param Optional[List[Path]] templates_dirs: A list of directories to load templates from using a
        :class:`nunavut.jinja.jinja2.FileSystemLoader`. If ``None`` no filesystem loader is created.
    :param bool followlinks: Argument passed on to the :class:`nunavut.jinja.jinja2.FileSystemLoader` instance.
    :param Optional[str] package_name_for_templates: The name of the package to load templates from. If ``None``
        then no :class:`nunavut.jinja.jinja2.PackageLoader` is created.
    :param str builtin_template_path: The name of the package under the ``package_name_for_templates`` package to load
        templates from. This is ignored if ``package_name_for_templates`` is None.
    :param Any kwargs: Arguments forwarded to the :class:`jinja.jinja2.BaseLoader`.
    """

    def __init__(
        self,
        templates_dirs: typing.Optional[typing.List[pathlib.Path]] = None,
        followlinks: bool = False,
        package_name_for_templates: typing.Optional[str] = None,
        builtin_template_path: str = DEFAULT_TEMPLATE_PATH,
        **kwargs: typing.Any
    ):
        super().__init__(**kwargs)
        self._type_to_template_lookup_cache = dict()  # type: typing.Dict[pydsdl.Any, pathlib.Path]
        self._templates_package_name = None  # type: typing.Optional[str]

        if templates_dirs is not None:
            for templates_dir_item in templates_dirs:
                if not pathlib.Path(templates_dir_item).exists:
                    raise ValueError("Templates directory {} did not exist?".format(templates_dir_item))
            logger.info("Loading templates from file system at {}".format(templates_dirs))
            self._fsloader = FileSystemLoader((str(d) for d in templates_dirs), followlinks=followlinks)
        else:
            self._fsloader = None

        if package_name_for_templates is not None:
            logger.info("Loading templates from package {}.{}".format(builtin_template_path, builtin_template_path))
            self._package_loader = PackageLoader(package_name_for_templates, package_path=builtin_template_path)
            self._templates_package_name = "{}.{}".format(package_name_for_templates, builtin_template_path)
        else:
            self._package_loader = None

    def get_source(
        self, environment: Environment, template: str
    ) -> typing.Tuple[typing.Any, str, typing.Callable[..., bool]]:
        if self._fsloader is not None:
            try:
                return typing.cast(
                    typing.Tuple[typing.Any, str, typing.Callable[..., bool]],
                    self._fsloader.get_source(environment, template),
                )
            except TemplateNotFound:
                if self._package_loader is None:
                    raise
        if self._package_loader is not None:
            return typing.cast(
                typing.Tuple[typing.Any, str, typing.Callable[..., bool]],
                self._package_loader.get_source(environment, template),
            )
        raise TemplateNotFound(template)

    def list_templates(self) -> typing.Iterable[str]:
        """
        Override of :meth:`BaseLoader.list_templates` that returns an aggregate of the filesystem loader and
        package loader templates.

        :return: A list of templates names (i.e. file stems) found by this Generator object.

        .. invisible-code-block: python

            from nunavut.jinja.loaders import DSDLTemplateLoader, TEMPLATE_SUFFIX

            template_loaders = DSDLTemplateLoader(package_name_for_templates='nunavut.lang.c')

            templates = template_loaders.list_templates()

            structure_type = None

            for template in templates:
                if template == 'StructureType' + TEMPLATE_SUFFIX:
                    structure_type = template

            assert structure_type is not None

        """
        files = []
        if self._fsloader is not None:
            files += self._filter_template_list_by_suffix(self._fsloader.list_templates())
        if self._package_loader is not None:
            files += self._filter_template_list_by_suffix(self._package_loader.list_templates())

        return files

    def get_template_sets(self) -> typing.List[typing.Tuple[str, str, typing.Tuple[int, int, int]]]:
        template_sets = []  # type: typing.List[typing.Tuple[str, str, typing.Tuple[int, int, int]]]
        if self._templates_package_name is not None:
            vr = VersionReader(self._templates_package_name)
            template_sets.append(("package", self._templates_package_name, vr.version))
        return template_sets

    def get_templates(self) -> typing.Iterable[pathlib.Path]:
        """
        Enumerate all templates found in the templates path.
        :data:`~TEMPLATE_SUFFIX` as the suffix for the filename. This method differs from the :class:`BaseLoader`
        override of :meth:`BaseLoader.list_templates` in that it returns paths instead of just file name stems.

        :return: A list of paths to all templates found by this Generator object.

        .. invisible-code-block: python

            from nunavut.jinja.loaders import DSDLTemplateLoader, TEMPLATE_SUFFIX

            template_loaders = DSDLTemplateLoader(package_name_for_templates='nunavut.lang.c')

            templates = template_loaders.get_templates()

            structure_type = None

            for template in templates:
                if template.stem == 'StructureType':
                    structure_type = template

            assert structure_type is not None
            assert structure_type.suffix == TEMPLATE_SUFFIX
            assert structure_type.exists()

        """
        files = set()
        if self._fsloader is not None:
            for template_dir in self._fsloader.searchpath:
                for template in pathlib.Path(template_dir).glob("**/*{}".format(TEMPLATE_SUFFIX)):
                    files.add(template)
        if self._package_loader is not None and self._templates_package_name is not None:
            templates_module = importlib.import_module(self._templates_package_name)
            spec_perhaps = templates_module.__spec__
            file_perhaps = None  # type: typing.Optional[str]
            if spec_perhaps is not None:
                file_perhaps = spec_perhaps.origin
            if file_perhaps is None or file_perhaps == "builtin":
                raise RuntimeError("Unknown template package origin?")
            templates_base_path = pathlib.Path(file_perhaps).parent
            for t in self._package_loader.list_templates():
                files.add(templates_base_path / pathlib.Path(t))
        return sorted(files)

    def type_to_template(self, value_type: typing.Type) -> typing.Optional[pathlib.Path]:
        """
        Given a type object, return a template used to generate code for the type.

        :return: a template or None if no template could be found for the given type.

        .. invisible-code-block: python
            from nunavut.jinja.loaders import DSDLTemplateLoader
            import pydsdl

            l = DSDLTemplateLoader(package_name_for_templates='nunavut.lang.py')
            template_name = l.type_to_template(pydsdl.StructureType)

            assert template_name is not None
            assert template_name.name == 'Any.j2'

        """
        template_path = None
        if self._fsloader is not None:
            filtered_templates = self._filter_template_list_by_suffix(self._fsloader.list_templates())
            template_path = self._type_to_template_internal(
                value_type, dict(map(lambda x: (pathlib.Path(x).stem, pathlib.Path(x)), filtered_templates))
            )
        if template_path is None and self._package_loader is not None:
            filtered_templates = self._filter_template_list_by_suffix(self._package_loader.list_templates())
            template_path = self._type_to_template_internal(
                value_type, dict(map(lambda x: (pathlib.Path(x).stem, pathlib.Path(x)), filtered_templates))
            )

        return template_path

    # +----------------------------------------------------------------------------------------------------------------+
    # | PRIVATE
    # +----------------------------------------------------------------------------------------------------------------+
    @staticmethod
    def _filter_template_list_by_suffix(files: typing.List[str]) -> typing.List[str]:
        return [f for f in files if (pathlib.Path(f).suffix == TEMPLATE_SUFFIX)]

    def _type_to_template_internal(
        self, value_type: typing.Type, templates: typing.Mapping[str, pathlib.Path]
    ) -> typing.Optional[pathlib.Path]:
        search_queue = collections.deque()  # type: typing.Deque[typing.Any]
        discovered = set()  # type: typing.Set[typing.Any]
        search_queue.appendleft(value_type)
        template_path = None

        while len(search_queue) > 0:
            current_search_type = search_queue.pop()
            try:
                template_path = self._type_to_template_lookup_cache[current_search_type]
                break
            except KeyError:
                pass

            try:
                logging.debug(
                    "NunavutTemplateLoader.type_to_template for {}: considering {}...".format(
                        value_type.__name__, current_search_type.__name__
                    )
                )
                template_path = templates[current_search_type.__name__]
                self._type_to_template_lookup_cache[current_search_type] = template_path
                break
            except KeyError:
                for base_type in current_search_type.__bases__:
                    if base_type != object and base_type not in discovered:
                        search_queue.appendleft(base_type)
                        discovered.add(current_search_type)

        return template_path
